import logging
import time

import attr
import pytest
from pytestskipmarkers.utils import platform
from saltfactories.utils import random_string

# This `pytest.importorskip` here actually works because this module
# is imported into test modules, otherwise, the skipping would just fail
pytest.importorskip("docker")
import docker.errors  # isort:skip  pylint: disable=3rd-party-module-not-gated

log = logging.getLogger(__name__)


@attr.s(kw_only=True, slots=True)
class MySQLImage:
    name = attr.ib()
    tag = attr.ib()
    container_id = attr.ib()

    def __str__(self):
        return f"{self.name}:{self.tag}"


@attr.s(kw_only=True, slots=True)
class MySQLCombo:
    mysql_name = attr.ib()
    mysql_version = attr.ib()
    mysql_port = attr.ib(default=None)
    mysql_host = attr.ib(default="%")
    mysql_user = attr.ib()
    mysql_passwd = attr.ib()
    mysql_database = attr.ib(default=None)
    mysql_root_user = attr.ib(default="root")
    mysql_root_passwd = attr.ib()
    container = attr.ib(default=None)
    container_id = attr.ib()

    @container_id.default
    def _default_container_id(self):
        return random_string(
            "{}-{}-".format(
                self.mysql_name.replace("/", "-"),
                self.mysql_version,
            )
        )

    @mysql_root_passwd.default
    def _default_mysql_root_user_passwd(self):
        return self.mysql_passwd

    def get_credentials(self, **kwargs):
        return {
            "connection_user": kwargs.get("connection_user") or self.mysql_root_user,
            "connection_pass": kwargs.get("connection_pass") or self.mysql_root_passwd,
            "connection_db": kwargs.get("connection_db") or "mysql",
            "connection_port": kwargs.get("connection_port") or self.mysql_port,
        }


def get_test_versions():
    test_versions = []
    name = "mysql-server"
    for version in ("5.5", "5.6", "5.7", "8.0"):
        test_versions.append(
            MySQLImage(
                name=name,
                tag=version,
                container_id=random_string(f"mysql-{version}-"),
            )
        )
    name = "mariadb"
    for version in ("10.3", "10.4", "10.5"):
        test_versions.append(
            MySQLImage(
                name=name,
                tag=version,
                container_id=random_string(f"mariadb-{version}-"),
            )
        )
    name = "percona"
    for version in ("5.6", "5.7", "8.0"):
        test_versions.append(
            MySQLImage(
                name=name,
                tag=version,
                container_id=random_string(f"percona-{version}-"),
            )
        )
    return test_versions


def get_test_version_id(value):
    return f"container={value}"


@pytest.fixture(scope="module", params=get_test_versions(), ids=get_test_version_id)
def mysql_image(request):
    return request.param


@pytest.fixture(scope="module")
def create_mysql_combo(mysql_image):
    if platform.is_fips_enabled():
        if mysql_image.name in ("mysql-server", "percona") and mysql_image.tag == "8.0":
            pytest.skip(f"These tests fail on {mysql_image.name}:{mysql_image.tag}")

    return MySQLCombo(
        mysql_name=mysql_image.name,
        mysql_version=mysql_image.tag,
        mysql_user="salt-mysql-user",
        mysql_passwd="Pa55w0rd!",
        container_id=mysql_image.container_id,
    )


@pytest.fixture(scope="module")
def mysql_combo(create_mysql_combo):
    return create_mysql_combo


def check_container_started(timeout_at, container, combo):
    sleeptime = 0.5
    while time.time() <= timeout_at:
        try:
            if not container.is_running():
                log.warning("%s is no longer running", container)
                return False
            ret = container.run(
                "mysql",
                f"--user={combo.mysql_user}",
                f"--password={combo.mysql_passwd}",
                "-e",
                "SELECT 1",
            )
            if ret.returncode == 0:
                break
        except docker.errors.APIError:
            log.exception("Failed to run start check")
        time.sleep(sleeptime)
        sleeptime *= 2
    else:
        return False
    time.sleep(0.5)
    return True


def set_container_name_before_start(container):
    """
    This is useful if the container has to be restared and the old
    container, under the same name was left running, but in a bad shape.
    """
    container.name = random_string("{}-".format(container.name.rsplit("-", 1)[0]))
    container.display_name = None
    return container


@pytest.fixture(scope="module")
def mysql_container(salt_factories, mysql_combo):

    container_environment = {
        "MYSQL_ROOT_PASSWORD": mysql_combo.mysql_passwd,
        "MYSQL_ROOT_HOST": mysql_combo.mysql_host,
        "MYSQL_USER": mysql_combo.mysql_user,
        "MYSQL_PASSWORD": mysql_combo.mysql_passwd,
    }
    if mysql_combo.mysql_database:
        container_environment["MYSQL_DATABASE"] = mysql_combo.mysql_database

    container = salt_factories.get_container(
        mysql_combo.container_id,
        "ghcr.io/saltstack/salt-ci-containers/{}:{}".format(
            mysql_combo.mysql_name, mysql_combo.mysql_version
        ),
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
        container_run_kwargs={
            "ports": {"3306/tcp": None},
            "environment": container_environment,
        },
    )
    container.before_start(set_container_name_before_start, container)
    container.container_start_check(check_container_started, container, mysql_combo)
    with container.started():
        mysql_combo.container = container
        mysql_combo.mysql_port = container.get_host_port_binding(
            3306, protocol="tcp", ipv6=False
        )
        yield mysql_combo



"""
Manage IPMI devices over LAN
============================

The following configuration defaults can be defined in the
minion, master config or pillar:

.. code-block:: yaml

    ipmi.config:
        api_host: 127.0.0.1
        api_user: admin
        api_pass: apassword
        api_port: 623
        api_kg: None

Every call can override the config defaults:

.. code-block:: yaml

    ensure myipmi system is set to network boot:
        ipmi.boot_device:
            - name: network
            - api_host: myipmi.hostname.com
            - api_user: root
            - api_pass: apassword
            - api_kg: None

    ensure myipmi system is powered on:
        ipmi.power:
            - name: boot
            - api_host: myipmi.hostname.com
            - api_user: root
            - api_pass: apassword
"""


def __virtual__():
    IMPORT_ERR = None
    try:
        # pylint: disable=unused-import
        from pyghmi.ipmi import command  # nosec

        # pylint: enable=unused-import
    except ImportError as exc:
        IMPORT_ERR = str(exc)
    return (IMPORT_ERR is None, IMPORT_ERR)


def boot_device(name="default", **kwargs):
    """
    Request power state change

    name = ``default``
        * network -- Request network boot
        * hd -- Boot from hard drive
        * safe -- Boot from hard drive, requesting 'safe mode'
        * optical -- boot from CD/DVD/BD drive
        * setup -- Boot into setup utility
        * default -- remove any IPMI directed boot device request

    kwargs
        - api_host=localhost
        - api_user=admin
        - api_pass=
        - api_port=623
        - api_kg=None
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    org = __salt__["ipmi.get_bootdev"](**kwargs)
    if "bootdev" in org:
        org = org["bootdev"]

    if org == name:
        ret["result"] = True
        ret["comment"] = "system already in this state"
        return ret

    if __opts__["test"]:
        ret["comment"] = "would change boot device"
        ret["result"] = None
        ret["changes"] = {"old": org, "new": name}
        return ret

    outdddd = __salt__["ipmi.set_bootdev"](bootdev=name, **kwargs)
    ret["comment"] = "changed boot device"
    ret["result"] = True
    ret["changes"] = {"old": org, "new": name}
    return ret


def power(name="power_on", wait=300, **kwargs):
    """
    Request power state change

    name
        Ensure power state one of:
            * power_on -- system turn on
            * power_off -- system turn off (without waiting for OS)
            * shutdown -- request OS proper shutdown
            * reset -- reset (without waiting for OS)
            * boot -- If system is off, then 'on', else 'reset'

    wait
        wait X seconds for the job to complete before forcing.
        (defaults to 300 seconds)

    kwargs
        - api_host=localhost
        - api_user=admin
        - api_pass=
        - api_port=623
        - api_kg=None
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    org = __salt__["ipmi.get_power"](**kwargs)

    state_map = {
        "off": "off",
        "on": "on",
        "power_off": "off",
        "power_on": "on",
        "shutdown": "off",
        "reset": "na",
        "boot": "na",
    }

    if org == state_map[name]:
        ret["result"] = True
        ret["comment"] = "system already in this state"
        return ret

    if __opts__["test"]:
        ret["comment"] = f"would power: {name} system"
        ret["result"] = None
        ret["changes"] = {"old": org, "new": name}
        return ret

    outdddd = __salt__["ipmi.set_power"](name, wait=wait, **kwargs)
    ret["comment"] = "changed system power"
    ret["result"] = True
    ret["changes"] = {"old": org, "new": name}
    return ret


def user_present(
    name,
    uid,
    password,
    channel=14,
    callback=False,
    link_auth=True,
    ipmi_msg=True,
    privilege_level="administrator",
    **kwargs,
):
    """
    Ensure IPMI user and user privileges.

    name
        name of user (limit 16 bytes)

    uid
        user id number (1 to 7)

    password
        user password (limit 16 bytes)

    channel
        ipmi channel defaults to 14 for auto

    callback
        User Restricted to Callback

        False = User Privilege Limit is determined by the User Privilege Limit
            parameter privilege_level, for both callback and non-callback connections.

        True  = User Privilege Limit is determined by the privilege_level
            parameter for callback connections, but is restricted to Callback
            level for non-callback connections. Thus, a user can only initiate
            a Callback when they 'call in' to the BMC, but once the callback
            connection has been made, the user could potentially establish a
            session as an Operator.

    link_auth
        User Link authentication
        True/False
        user name and password information will be used for link
        authentication, e.g. PPP CHAP) for the given channel. Link
        authentication itself is a global setting for the channel and is
        enabled/disabled via the serial/modem configuration parameters.

    ipmi_msg
        User IPMI Messaging
        True/False
        user name and password information will be used for IPMI
        Messaging. In this case, 'IPMI Messaging' refers to the ability to
        execute generic IPMI commands that are not associated with a
        particular payload type. For example, if IPMI Messaging is disabled for
        a user, but that user is enabled for activating the SOL
        payload type, then IPMI commands associated with SOL and session
        management, such as Get SOL Configuration Parameters and Close Session
        are available, but generic IPMI commands such as Get SEL Time are
        unavailable.)
        ipmi_msg

    privilege_level
        * callback
        * user
        * operator
        * administrator
        * proprietary
        * no_access

    kwargs
        - api_host=localhost
        - api_user=admin
        - api_pass=
        - api_port=623
        - api_kg=None
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    org_user = __salt__["ipmi.get_user"](uid=uid, channel=channel, **kwargs)

    change = False
    if org_user["access"]["callback"] != callback:
        change = True
    if org_user["access"]["link_auth"] != link_auth:
        change = True
    if org_user["access"]["ipmi_msg"] != ipmi_msg:
        change = True
    if org_user["access"]["privilege_level"] != privilege_level:
        change = True
    if (
        __salt__["ipmi.set_user_password"](
            uid, mode="test_password", password=password, **kwargs
        )
        is False
    ):
        change = True

    if change is False:
        ret["result"] = True
        ret["comment"] = "user already present"
        return ret

    if __opts__["test"]:
        ret["comment"] = "would (re)create user"
        ret["result"] = None
        ret["changes"] = {"old": org_user, "new": name}
        return ret

    __salt__["ipmi.ensure_user"](
        uid,
        name,
        password,
        channel,
        callback,
        link_auth,
        ipmi_msg,
        privilege_level,
        **kwargs,
    )
    current_user = __salt__["ipmi.get_user"](uid=uid, channel=channel, **kwargs)
    ret["comment"] = "(re)created user"
    ret["result"] = True
    ret["changes"] = {"old": org_user, "new": current_user}
    return ret


def user_absent(name, channel=14, **kwargs):
    """
    Remove user
    Delete all user (uid) records having the matching name.

    name
        string name of user to delete

    channel
        channel to remove user access from defaults to 14 for auto.

    kwargs
        - api_host=localhost
        - api_user=admin
        - api_pass=
        - api_port=623
        - api_kg=None
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    user_id_list = __salt__["ipmi.get_name_uids"](name, channel, **kwargs)

    if len(user_id_list) == 0:
        ret["result"] = True
        ret["comment"] = "user already absent"
        return ret

    if __opts__["test"]:
        ret["comment"] = "would delete user(s)"
        ret["result"] = None
        ret["changes"] = {"delete": user_id_list}
        return ret

    for uid in user_id_list:
        __salt__["ipmi.delete_user"](uid, channel, **kwargs)

    ret["comment"] = "user(s) removed"
    ret["changes"] = {"old": user_id_list, "new": "None"}
    return ret



import copy
import logging

import pytest

import salt.output.highstate as highstate
import salt.utils.stringutils
from tests.support.mock import patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts.update({"color": False, "state_output_pct": True})
    return {highstate: {"__opts__": minion_opts}}


@pytest.fixture
def json_data():
    return {
        "data": {
            "master": {
                "salt_|-call_sleep_state_|-call_sleep_state_|-state": {
                    "__id__": "call_sleep_state",
                    "__jid__": "20170418153529810135",
                    "__run_num__": 0,
                    "__sls__": "orch.simple",
                    "changes": {
                        "out": "highstate",
                        "ret": {
                            "minion": {
                                "module_|-simple-ping_|-test.ping_|-run": {
                                    "__id__": "simple-ping",
                                    "__run_num__": 0,
                                    "__sls__": "simple-ping",
                                    "changes": {"ret": True},
                                    "comment": "Module function test.ping executed",
                                    "duration": 56.179,
                                    "name": "test.ping",
                                    "result": True,
                                    "start_time": "15:35:31.282099",
                                }
                            },
                            "sub_minion": {
                                "module_|-simple-ping_|-test.ping_|-run": {
                                    "__id__": "simple-ping",
                                    "__run_num__": 0,
                                    "__sls__": "simple-ping",
                                    "changes": {"ret": True},
                                    "comment": "Module function test.ping executed",
                                    "duration": 54.103,
                                    "name": "test.ping",
                                    "result": True,
                                    "start_time": "15:35:31.005606",
                                }
                            },
                        },
                    },
                    "comment": (
                        "States ran successfully. Updating sub_minion, minion."
                    ),
                    "duration": 1638.047,
                    "name": "call_sleep_state",
                    "result": True,
                    "start_time": "15:35:29.762657",
                },
                "salt_|-cmd_run_example_|-cmd.run_|-function": {
                    "__id__": "cmd_run_example",
                    "__jid__": "20200411195112288850",
                    "__run_num__": 1,
                    "__sls__": "orch.simple",
                    "changes": {
                        "out": "highstate",
                        "ret": {"minion": "file1\nfile2\nfile3"},
                    },
                    "comment": (
                        "Function ran successfully. Function cmd.run ran on minion."
                    ),
                    "duration": 412.397,
                    "name": "cmd.run",
                    "result": True,
                    "start_time": "21:51:12.185868",
                },
            }
        },
        "outputter": "highstate",
        "retcode": 0,
    }


@pytest.mark.parametrize("data", [None, {"return": None}, {"return": {"data": None}}])
def test_when_data_result_is_None_output_should_be_string_None(data):
    expected_output = "None"

    actual_output = highstate.output(data=data)

    assert actual_output == expected_output


def test_when_data_is_dict_with_return_key_and_return_value_has_data_key_and_data_dict_has_one_dict_element_with_jid_and_fun_keys_and_return_value_is_None_then_output_should_return_literal_None_string():

    expected_output = "None"
    data = {
        "return": {
            "data": {
                "foo bar quux fnord": {
                    "jid": "fnordy fnordy fnordy",
                    "fun": "fnordy fnordy fnord",
                    "return": {"data": None},
                },
            }
        },
    }

    actual_output = highstate.output(data=data)

    assert actual_output == expected_output


@pytest.mark.parametrize(
    "return_value",
    [42, "fnord"],
)
def test_when_data_is_dict_with_return_key_and_return_value_has_data_key_and_data_dict_has_one_dict_element_with_jid_and_fun_keys_and_return_value_is_int_or_str_that_value_should_be_returned(
    return_value,
):

    expected_output = return_value
    data = {
        "return": {
            "data": {
                "foo bar quux fnord": {
                    "jid": "fnordy fnordy fnordy",
                    "fun": "fnordy fnordy fnord",
                    "return": {"data": return_value},
                },
            }
        },
    }

    actual_output = highstate.output(data=data)

    assert actual_output == expected_output


def test_when_orchestrator_output_retcode_in_data_the_retcode_should_be_removed():
    data = {"something_master": None, "retcode": 42}

    actual_output = highstate.output(data)

    assert "retcode" not in data


def test_when_more_than_one_local_master_retcode_should_not_be_removed():
    expected_retcode = 42
    data = {
        "something_master": None,
        "another_master": None,
        "retcode": expected_retcode,
    }

    actual_output = highstate.output(data)

    assert data["retcode"] == expected_retcode


def test_pct_summary_output():
    data = {
        "data": {
            "master": {
                "salt_|-call_sleep_state_|-call_sleep_state_|-state": {
                    "__id__": "call_sleep_state",
                    "__jid__": "20170418153529810135",
                    "__run_num__": 0,
                    "__sls__": "orch.simple",
                    "changes": {
                        "out": "highstate",
                        "ret": {
                            "minion": {
                                "module_|-simple-ping_|-test.ping_|-run": {
                                    "__id__": "simple-ping",
                                    "__run_num__": 0,
                                    "__sls__": "simple-ping",
                                    "changes": {"ret": True},
                                    "comment": "Module function test.ping executed",
                                    "duration": 56.179,
                                    "name": "test.ping",
                                    "result": True,
                                    "start_time": "15:35:31.282099",
                                }
                            },
                            "sub_minion": {
                                "module_|-simple-ping_|-test.ping_|-run": {
                                    "__id__": "simple-ping",
                                    "__run_num__": 0,
                                    "__sls__": "simple-ping",
                                    "changes": {"ret": True},
                                    "comment": "Module function test.ping executed",
                                    "duration": 54.103,
                                    "name": "test.ping",
                                    "result": True,
                                    "start_time": "15:35:31.005606",
                                }
                            },
                        },
                    },
                    "comment": (
                        "States ran successfully. Updating sub_minion, minion."
                    ),
                    "duration": 1638.047,
                    "name": "call_sleep_state",
                    "result": True,
                    "start_time": "15:35:29.762657",
                },
                "salt_|-cmd_run_example_|-cmd.run_|-function": {
                    "__id__": "cmd_run_example",
                    "__jid__": "20200411195112288850",
                    "__run_num__": 1,
                    "__sls__": "orch.simple",
                    "changes": {
                        "out": "highstate",
                        "ret": {"minion": "file1\nfile2\nfile3"},
                    },
                    "comment": (
                        "Function ran successfully. Function cmd.run ran on minion."
                    ),
                    "duration": 412.397,
                    "name": "cmd.run",
                    "result": True,
                    "start_time": "21:51:12.185868",
                },
            }
        },
        "outputter": "highstate",
        "retcode": 0,
    }

    actual_output = highstate.output(data)
    assert "Succeeded: 1 (changed=1)" in actual_output
    assert "Failed:    0" in actual_output
    assert "Success %: 100.0" in actual_output
    assert "Failure %: 0.0" in actual_output
    assert "Total states run:     1" in actual_output
    assert "                  file2" in actual_output


def test__compress_ids():
    """
    Tests for expected data return for _compress_ids
    and proper formatting using the state_compress_ids option
    """
    # raw data entering the outputter
    data = {
        "local": {
            "cmd_|-mix-matched results_|-/bin/false_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 7,
                "__sls__": "compress_test",
                "changes": {"pid": 6554, "retcode": 1, "stderr": "", "stdout": ""},
                "comment": "Command " '"/bin/false" ' "run",
                "duration": 8.57,
                "name": "/bin/false",
                "result": False,
                "start_time": "15:38:22.666578",
            },
            "cmd_|-mix-matched results_|-/bin/true_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 6,
                "__sls__": "compress_test",
                "changes": {"pid": 6553, "retcode": 0, "stderr": "", "stdout": ""},
                "comment": "Command " '"/bin/true" ' "run",
                "duration": 7.728,
                "name": "/bin/true",
                "result": True,
                "start_time": "15:38:22.658452",
            },
            "cmd_|-mix-matched results_|-false_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 5,
                "__sls__": "compress_test",
                "changes": {"pid": 6552, "retcode": 1, "stderr": "", "stdout": ""},
                "comment": "Command " '"false" run',
                "duration": 7.832,
                "name": "false",
                "result": False,
                "start_time": "15:38:22.650225",
            },
            "cmd_|-mix-matched results_|-true_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 4,
                "__sls__": "compress_test",
                "changes": {"pid": 6551, "retcode": 0, "stderr": "", "stdout": ""},
                "comment": "Command " '"true" run',
                "duration": 8.538,
                "name": "true",
                "result": True,
                "start_time": "15:38:22.641293",
            },
            "file_|-one clean one changes_|-/tmp/changes_|-managed": {
                "__id__": "one clean one changes",
                "__run_num__": 13,
                "__sls__": "compress_test",
                "changes": {"diff": "New file"},
                "comment": "File /tmp/changes updated",
                "duration": 3.17,
                "name": "/tmp/changes",
                "result": True,
                "start_time": "15:38:22.703770",
            },
            "file_|-one clean one changes_|-/tmp/clean_|-managed": {
                "__id__": "one clean one changes",
                "__run_num__": 12,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "File /tmp/clean is in the correct state",
                "duration": 20.123,
                "name": "/tmp/clean",
                "result": True,
                "start_time": "15:38:22.683450",
            },
            "test_|-succeed clean_|-bar_|-succeed_without_changes": {
                "__id__": "succeed clean",
                "__run_num__": 11,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 0.759,
                "name": "bar",
                "result": True,
                "start_time": "15:38:22.678512",
            },
            "test_|-succeed clean_|-foo_|-succeed_without_changes": {
                "__id__": "succeed clean",
                "__run_num__": 10,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 0.676,
                "name": "foo",
                "result": True,
                "start_time": "15:38:22.677678",
            },
            "test_|-succeed clean_|-hello_|-succeed_without_changes": {
                "__id__": "succeed clean",
                "__run_num__": 8,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 1.071,
                "name": "hello",
                "result": True,
                "start_time": "15:38:22.675588",
            },
            "test_|-succeed clean_|-world_|-succeed_without_changes": {
                "__id__": "succeed clean",
                "__run_num__": 9,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 0.693,
                "name": "world",
                "result": True,
                "start_time": "15:38:22.676826",
            },
            "test_|-succeed with changes_|-bar_|-succeed_with_changes": {
                "__id__": "succeed with changes",
                "__run_num__": 3,
                "__sls__": "compress_test",
                "changes": {
                    "testing": {
                        "new": "Something pretended to change",
                        "old": "Unchanged",
                    }
                },
                "comment": "Success!",
                "duration": 0.829,
                "name": "bar",
                "result": True,
                "start_time": "15:38:22.639625",
            },
            "test_|-succeed with changes_|-foo_|-succeed_with_changes": {
                "__id__": "succeed with changes",
                "__run_num__": 2,
                "__sls__": "compress_test",
                "changes": {
                    "testing": {
                        "new": "Something pretended to change",
                        "old": "Unchanged",
                    }
                },
                "comment": "Success!",
                "duration": 0.739,
                "name": "foo",
                "result": True,
                "start_time": "15:38:22.638724",
            },
            "test_|-succeed with changes_|-hello_|-succeed_with_changes": {
                "__id__": "succeed with changes",
                "__run_num__": 0,
                "__sls__": "compress_test",
                "changes": {
                    "testing": {
                        "new": "Something pretended to change",
                        "old": "Unchanged",
                    }
                },
                "comment": "Success!",
                "duration": 0.812,
                "name": "hello",
                "result": True,
                "start_time": "15:38:22.636883",
            },
            "test_|-succeed with changes_|-world_|-succeed_with_changes": {
                "__id__": "succeed with changes",
                "__run_num__": 1,
                "__sls__": "compress_test",
                "changes": {
                    "testing": {
                        "new": "Something pretended to change",
                        "old": "Unchanged",
                    }
                },
                "comment": "Success!",
                "duration": 0.694,
                "name": "world",
                "result": True,
                "start_time": "15:38:22.637872",
            },
            "test_|-single clean_|-single_|-succeed_without_changes": {
                "__id__": "single clean",
                "__run_num__": 14,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 0.693,
                "name": "single",
                "result": True,
                "start_time": "15:38:22.676827",
            },
        }
    }
    # expected compressed raw data for outputter
    expected_output = {
        "local": {
            "cmd_|-mix-matched results (2)_|-state_compressed_compress_test_mix-matched results_False_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 5,
                "__sls__": "compress_test",
                "changes": {
                    "compressed changes": {
                        "/bin/false": {
                            "pid": 6554,
                            "retcode": 1,
                            "stderr": "",
                            "stdout": "",
                        },
                        "false": {
                            "pid": 6552,
                            "retcode": 1,
                            "stderr": "",
                            "stdout": "",
                        },
                    }
                },
                "comment": "Command " '"/bin/false" ' "run",
                "duration": 16.402,
                "name": "/bin/false",
                "result": False,
                "start_time": "15:38:22.650225",
            },
            "cmd_|-mix-matched results (2)_|-state_compressed_compress_test_mix-matched results_True_|-run": {
                "__id__": "mix-matched results",
                "__run_num__": 4,
                "__sls__": "compress_test",
                "changes": {
                    "compressed changes": {
                        "/bin/true": {
                            "pid": 6553,
                            "retcode": 0,
                            "stderr": "",
                            "stdout": "",
                        },
                        "true": {"pid": 6551, "retcode": 0, "stderr": "", "stdout": ""},
                    }
                },
                "comment": "Command " '"/bin/true" ' "run",
                "duration": 16.266,
                "name": "/bin/true",
                "result": True,
                "start_time": "15:38:22.641293",
            },
            "file_|-one clean one changes (2)_|-state_compressed_compress_test_one clean one changes_True_|-managed": {
                "__id__": "one clean one changes",
                "__run_num__": 12,
                "__sls__": "compress_test",
                "changes": {"diff": "New file"},
                "comment": "File /tmp/changes updated",
                "duration": 23.293,
                "name": "/tmp/changes",
                "result": True,
                "start_time": "15:38:22.683450",
            },
            "test_|-succeed clean (4)_|-state_compressed_compress_test_succeed clean_True_|-succeed_without_changes": {
                "__id__": "succeed clean",
                "__run_num__": 8,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 3.199,
                "name": "bar",
                "result": True,
                "start_time": "15:38:22.675588",
            },
            "test_|-succeed with changes (4)_|-state_compressed_compress_test_succeed with changes_True_|-succeed_with_changes": {
                "__id__": "succeed with changes",
                "__run_num__": 0,
                "__sls__": "compress_test",
                "changes": {
                    "compressed changes": {
                        "bar": {
                            "testing": {
                                "new": "Something pretended to change",
                                "old": "Unchanged",
                            }
                        },
                        "foo": {
                            "testing": {
                                "new": "Something pretended to change",
                                "old": "Unchanged",
                            }
                        },
                        "hello": {
                            "testing": {
                                "new": "Something pretended to change",
                                "old": "Unchanged",
                            }
                        },
                        "world": {
                            "testing": {
                                "new": "Something pretended to change",
                                "old": "Unchanged",
                            }
                        },
                    }
                },
                "comment": "Success!",
                "duration": 3.074,
                "name": "bar",
                "result": True,
                "start_time": "15:38:22.636883",
            },
            "test_|-single clean_|-single_|-succeed_without_changes": {
                "__id__": "single clean",
                "__run_num__": 14,
                "__sls__": "compress_test",
                "changes": {},
                "comment": "Success!",
                "duration": 0.693,
                "name": "single",
                "result": True,
                "start_time": "15:38:22.676827",
            },
        }
    }
    actual_output = highstate._compress_ids(data)

    # return properly compressed data
    assert actual_output == expected_output

    # check output text for formatting
    opts = copy.deepcopy(highstate.__opts__)
    opts["state_compress_ids"] = True
    with patch("salt.output.highstate.__opts__", opts, create=True):
        actual_output = highstate.output(data)
    assert "          ID: succeed with changes (4)" in actual_output
    assert (
        "        Name: state_compressed_compress_test_succeed with changes_True"
        in actual_output
    )
    assert "              compressed changes:" in actual_output
    assert "          ID: mix-matched results (2)" in actual_output
    assert (
        "        Name: state_compressed_compress_test_mix-matched results_True"
        in actual_output
    )
    assert (
        "        Name: state_compressed_compress_test_mix-matched results_False"
        in actual_output
    )
    assert "          ID: succeed clean (4)" in actual_output
    assert (
        "        Name: state_compressed_compress_test_succeed clean_True"
        in actual_output
    )
    assert "          ID: one clean one changes (2)" in actual_output
    assert (
        "        Name: state_compressed_compress_test_one clean one changes_True"
        in actual_output
    )
    assert "              diff:" in actual_output
    assert "Succeeded: 13 (changed=9)" in actual_output
    assert "Failed:     2" in actual_output
    assert "Success %: 86.67" in actual_output
    assert "Failure %: 13.33" in actual_output
    assert "Total states run:     15" in actual_output

    # pop out a __run_num__ to break the data
    data["local"]["cmd_|-mix-matched results_|-/bin/false_|-run"].pop("__run_num__")
    actual_output = highstate._compress_ids(data)

    # expecting return of original data to let the outputter figure it out
    assert actual_output == data


def test__compress_ids_not_dict():
    """
    Simple test for returning original malformed data
    to let the outputter figure it out.
    """
    data = ["malformed"]
    actual_output = highstate._compress_ids(data)
    assert actual_output == data


def test__compress_ids_multiple_module_functions():
    """
    Tests for expected data return for _compress_ids
    when using multiple mod.fun combos under the ID
    """
    # raw data entering the outputter
    data = {
        "local": {
            "cmd_|-try_this_|-echo 'hi'_|-run": {
                "__id__": "try_this",
                "__run_num__": 1,
                "__sls__": "wayne",
                "changes": {"pid": 32615, "retcode": 0, "stderr": "", "stdout": "hi"},
                "comment": 'Command "echo ' "'hi'\" run",
                "duration": 8.218,
                "name": "echo 'hi'",
                "result": True,
                "start_time": "23:43:25.715842",
            },
            "test_|-try_this_|-asdf_|-nop": {
                "__id__": "try_this",
                "__run_num__": 0,
                "__sls__": "wayne",
                "changes": {},
                "comment": "Success!",
                "duration": 0.906,
                "name": "asdf",
                "result": True,
                "start_time": "23:43:25.714010",
            },
        }
    }

    # check output text for formatting
    opts = copy.deepcopy(highstate.__opts__)
    opts["state_compress_ids"] = True
    with patch("salt.output.highstate.__opts__", opts, create=True):
        actual_output = highstate.output(data)

    # if we only cared about the ID/SLS/Result combo, this would be 4 not 2
    assert "Succeeded: 2 (changed=1)" in actual_output
    assert "Failed:    0" in actual_output
    assert "Total states run:     2" in actual_output


def test_parallel_summary_output():
    data = {
        "local": {
            "test_|-barrier_|-barrier_|-nop": {
                "name": "barrier",
                "changes": {},
                "result": True,
                "comment": "Success!",
                "__sls__": "test.49273",
                "__run_num__": 0,
                "start_time": "15:11:31.459770",
                "duration": 0.698,
                "__id__": "barrier",
            },
            "cmd_|-blah-1_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524313, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 1,
                "start_time": "15:11:31.496410",
                "duration": 10007.711,
                "__id__": "blah-1",
                "__parallel__": True,
            },
            "cmd_|-blah-2_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524315, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 2,
                "start_time": "15:11:31.537001",
                "duration": 10007.264,
                "__id__": "blah-2",
                "__parallel__": True,
            },
            "cmd_|-blah-3_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524317, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 3,
                "start_time": "15:11:31.577076",
                "duration": 10008.646,
                "__id__": "blah-3",
                "__parallel__": True,
            },
            "test_|-barrier2_|-barrier2_|-nop": {
                "name": "barrier2",
                "changes": {},
                "result": True,
                "comment": "Success!",
                "__sls__": "test.49273",
                "__run_num__": 4,
                "start_time": "15:11:41.619874",
                "duration": 0.762,
                "__id__": "barrier2",
            },
        }
    }

    actual_output = highstate.output(data)

    assert "Succeeded: 5 (changed=3)" in actual_output
    assert "Failed:    0" in actual_output
    assert "Total states run:     5" in actual_output

    # The three main states were run in parallel and slept for
    # 10 seconds each so the total run time should around 10 seconds
    assert "Total run time:  10.010 s" in actual_output


def test_summary_output():
    data = {
        "local": {
            "test_|-barrier_|-barrier_|-nop": {
                "name": "barrier",
                "changes": {},
                "result": True,
                "comment": "Success!",
                "__sls__": "test.49273",
                "__run_num__": 0,
                "start_time": "15:11:31.459770",
                "duration": 0.698,
                "__id__": "barrier",
            },
            "cmd_|-blah-1_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524313, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 1,
                "start_time": "15:11:31.496410",
                "duration": 10007.711,
                "__id__": "blah-1",
            },
            "cmd_|-blah-2_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524315, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 2,
                "start_time": "15:11:31.537001",
                "duration": 10007.264,
                "__id__": "blah-2",
            },
            "cmd_|-blah-3_|-sleep 10_|-run": {
                "name": "sleep 10",
                "result": True,
                "changes": {"pid": 524317, "retcode": 0, "stdout": "", "stderr": ""},
                "comment": 'Command "sleep 10" run',
                "__sls__": "test.49273",
                "__run_num__": 3,
                "start_time": "15:11:31.577076",
                "duration": 10008.646,
                "__id__": "blah-3",
            },
            "test_|-barrier2_|-barrier2_|-nop": {
                "name": "barrier2",
                "changes": {},
                "result": True,
                "comment": "Success!",
                "__sls__": "test.49273",
                "__run_num__": 4,
                "start_time": "15:11:41.619874",
                "duration": 0.762,
                "__id__": "barrier2",
            },
        }
    }

    actual_output = highstate.output(data)

    assert "Succeeded: 5 (changed=3)" in actual_output
    assert "Failed:    0" in actual_output
    assert "Total states run:     5" in actual_output

    # The three main states were not run in parallel and slept for
    # 10 seconds each so the total run time should around 30 seconds
    assert "Total run time:  30.025 s" in actual_output


def test_default_output(json_data):
    ret = highstate.output(json_data)
    assert "Succeeded: 1 (changed=1)" in ret
    assert "Failed:    0" in ret
    assert "Total states run:     1" in ret
    assert "                  file2" in ret


def test_output_comment_is_not_unicode(json_data):
    entry = None
    for key in (
        "data",
        "master",
        "salt_|-call_sleep_state_|-call_sleep_state_|-state",
        "changes",
        "ret",
        "minion",
        "module_|-simple-ping_|-test.ping_|-run",
    ):
        if entry is None:
            entry = json_data[key]
            continue
        entry = entry[key]
    entry["comment"] = salt.utils.stringutils.to_bytes(entry["comment"])
    ret = highstate.output(json_data)
    assert "Succeeded: 1 (changed=1)" in ret
    assert "Failed:    0" in ret
    assert "Total states run:     1" in ret
    assert "                  file2" in ret


def test_nested_output():
    nested_data = {
        "outputter": "highstate",
        "data": {
            "local_master": {
                "salt_|-nested_|-state.orchestrate_|-runner": {
                    "comment": "Runner function 'state.orchestrate' executed.",
                    "name": "state.orchestrate",
                    "start_time": "09:22:53.158742",
                    "result": True,
                    "duration": 980.694,
                    "__run_num__": 0,
                    "__jid__": "20180326092253538853",
                    "__sls__": "orch.test.nested",
                    "changes": {
                        "return": {
                            "outputter": "highstate",
                            "data": {
                                "local_master": {
                                    "test_|-always-passes-with-changes_|-oinaosf_|-succeed_with_changes": {
                                        "comment": "Success!",
                                        "name": "oinaosf",
                                        "start_time": "09:22:54.128415",
                                        "result": True,
                                        "duration": 0.437,
                                        "__run_num__": 0,
                                        "__sls__": "orch.test.changes",
                                        "changes": {
                                            "testing": {
                                                "new": (
                                                    "Something pretended to change"
                                                ),
                                                "old": "Unchanged",
                                            }
                                        },
                                        "__id__": "always-passes-with-changes",
                                    },
                                    "test_|-always-passes_|-fasdfasddfasdfoo_|-succeed_without_changes": {
                                        "comment": "Success!",
                                        "name": "fasdfasddfasdfoo",
                                        "start_time": "09:22:54.128986",
                                        "result": True,
                                        "duration": 0.25,
                                        "__run_num__": 1,
                                        "__sls__": "orch.test.changes",
                                        "changes": {},
                                        "__id__": "always-passes",
                                    },
                                }
                            },
                            "retcode": 0,
                        }
                    },
                    "__id__": "nested",
                }
            }
        },
        "retcode": 0,
    }

    ret = highstate.output(nested_data)
    assert "Succeeded: 1 (changed=1)" in ret
    assert "Failed:    0" in ret
    assert "Total states run:     1" in ret

    # the whitespace is relevant in this case, it is testing that it is nested
    assert "                        ID: always-passes-with-changes" in ret
    assert "                   Started: 09:22:54.128415" in ret
    assert "              Succeeded: 2 (changed=1)" in ret
    assert "              Failed:    0" in ret
    assert "              Total states run:     2" in ret



"""
Integration tests for salt-ssh logging
"""

import logging
import time

import pytest
from saltfactories.utils import random_string

from tests.support.helpers import Keys

pytest.importorskip("docker")


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]


@pytest.fixture(scope="module")
def ssh_auth():
    return random_string("sshpassword"), "app-admin"


@pytest.fixture(scope="module")
def ssh_keys(tmp_path_factory):
    """
    Temporary ssh key fixture
    """
    with Keys(tmp_path_factory) as keys:
        yield keys


@pytest.fixture(scope="module")
def ssh_docker_container(salt_factories, ssh_keys, ssh_auth):
    """
    Temporary docker container with python 3.6 and ssh enabled
    """
    ssh_pass, ssh_user = ssh_auth
    container = salt_factories.get_container(
        random_string("ssh-py_versions-"),
        "ghcr.io/saltstack/salt-ci-containers/ssh-minion:latest",
        container_run_kwargs={
            "ports": {
                "22/tcp": None,
            },
            "environment": {
                "SSH_USER": ssh_user,
                "SSH_AUTHORIZED_KEYS": ssh_keys.pub,
                "SSH_USER_PASSWORD": ssh_pass,
            },
            "cap_add": ["IPC_LOCK"],
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with container.started() as factory:
        factory.run(f"echo {ssh_pass} | passwd {ssh_user} --stdin")
        yield factory


@pytest.fixture(scope="module")
def ssh_port(ssh_docker_container):
    return ssh_docker_container.get_host_port_binding(22, protocol="tcp")


@pytest.fixture(scope="module")
def salt_ssh_roster_file(ssh_port, ssh_keys, salt_master, ssh_auth, known_hosts_file):
    """
    Temporary roster for ssh docker container
    """
    ssh_pass, ssh_user = ssh_auth
    roster = f"""
    pyvertest:
      host: localhost
      user: {ssh_user}
      port: {ssh_port}
      passwd: {ssh_pass}
      sudo: True
      sudo_user: root
      tty: True
      ssh_options:
        - UserKnownHostsFile={known_hosts_file}
    """
    with pytest.helpers.temp_file(
        "py_versions_roster", roster, salt_master.config_dir
    ) as roster_file:
        yield roster_file


@pytest.fixture(scope="module")
def salt_ssh_cli(salt_master, salt_ssh_roster_file, ssh_keys, ssh_docker_container):
    assert salt_master.is_running()
    assert ssh_docker_container.is_running()
    return salt_master.salt_ssh_cli(
        timeout=180,
        roster_file=salt_ssh_roster_file,
        target_host="localhost",
        ssh_user="app-admin",
        base_script_args=["--ignore-host-keys"],
    )


def test_log_password(salt_ssh_cli, caplog, ssh_auth):
    """
    Test to ensure password is not logged when
    using sudo and a password
    """
    ssh_pass, _ = ssh_auth
    with caplog.at_level(logging.TRACE):
        ret = salt_ssh_cli.run("--log-level=trace", "test.ping", minion_tgt="pyvertest")
    if "kex_exchange_identification" in ret.stdout:
        pytest.skip("Container closed ssh connection, skipping for now")
    try:
        assert ret.returncode == 0
    except AssertionError:
        time.sleep(5)
        with caplog.at_level(logging.TRACE):
            ret = salt_ssh_cli.run(
                "--log-level=trace", "test.ping", minion_tgt="pyvertest"
            )
        if "kex_exchange_identification" in ret.stdout:
            pytest.skip("Container closed ssh connection, skipping for now")
        assert ret.returncode == 0
    assert ssh_pass not in caplog.text
    assert ret.data is True



"""
Execute a command and read the output as JSON. The JSON data is then directly overlaid onto the minion's Pillar data.
"""

import logging

import salt.utils.json

# Don't "fix" the above docstring to put it on two lines, as the sphinx
# autosummary pulls only the first line for its description.


# Set up logging
log = logging.getLogger(__name__)


def ext_pillar(
    minion_id, pillar, command  # pylint: disable=W0613  # pylint: disable=W0613
):
    """
    Execute a command and read the output as JSON
    """
    try:
        command = command.replace("%s", minion_id)
        return salt.utils.json.loads(__salt__["cmd.run"](command))
    except Exception:  # pylint: disable=broad-except
        log.critical("JSON data from %s failed to parse", command)
        return {}



"""
Connection module for Amazon ELB

.. versionadded:: 2014.7.0

:configuration: This module accepts explicit elb credentials but can also utilize
    IAM roles assigned to the instance through Instance Profiles. Dynamic
    credentials are then automatically obtained from AWS API and no further
    configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        elb.keyid: GKTADJGHEIQSXMKKRBJ08H
        elb.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        elb.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto >= 2.33.0
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602

import logging
import time
from collections import OrderedDict

import salt.utils.json
import salt.utils.versions

try:
    import boto
    import boto.ec2  # pylint: enable=unused-import
    from boto.ec2.elb import HealthCheck
    from boto.ec2.elb.attributes import (
        AccessLogAttribute,
        ConnectionDrainingAttribute,
        ConnectionSettingAttribute,
        CrossZoneLoadBalancingAttribute,
    )

    logging.getLogger("boto").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    # connection settings were added in 2.33.0
    has_boto_reqs = salt.utils.versions.check_boto_reqs(
        boto_ver="2.33.0", check_boto3=False
    )
    if has_boto_reqs is True:
        __utils__["boto.assign_funcs"](__name__, "elb", module="ec2.elb", pack=__salt__)
    return has_boto_reqs


def exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if an ELB exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.exists myelb region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        elb = conn.get_all_load_balancers(load_balancer_names=[name])
        if elb:
            return True
        else:
            log.debug("The load balancer does not exist in region %s", region)
            return False
    except boto.exception.BotoServerError as error:
        log.warning(error)
        return False


def get_all_elbs(region=None, key=None, keyid=None, profile=None):
    """
    Return all load balancers associated with an account

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.get_all_elbs region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        return [e for e in conn.get_all_load_balancers()]
    except boto.exception.BotoServerError as error:
        log.warning(error)
        return []


def list_elbs(region=None, key=None, keyid=None, profile=None):
    """
    Return names of all load balancers associated with an account

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.list_elbs region=us-east-1
    """

    return [
        e.name
        for e in get_all_elbs(region=region, key=key, keyid=keyid, profile=profile)
    ]


def get_elb_config(name, region=None, key=None, keyid=None, profile=None):
    """
    Get an ELB configuration.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.exists myelb region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30

    while retries:
        try:
            lb = conn.get_all_load_balancers(load_balancer_names=[name])
            lb = lb[0]
            ret = {}
            ret["availability_zones"] = lb.availability_zones
            listeners = []
            for _listener in lb.listeners:
                listener_dict = {}
                listener_dict["elb_port"] = _listener.load_balancer_port
                listener_dict["elb_protocol"] = _listener.protocol
                listener_dict["instance_port"] = _listener.instance_port
                listener_dict["instance_protocol"] = _listener.instance_protocol
                listener_dict["policies"] = _listener.policy_names
                if _listener.ssl_certificate_id:
                    listener_dict["certificate"] = _listener.ssl_certificate_id
                listeners.append(listener_dict)
            ret["listeners"] = listeners
            backends = []
            for _backend in lb.backends:
                bs_dict = {}
                bs_dict["instance_port"] = _backend.instance_port
                bs_dict["policies"] = [p.policy_name for p in _backend.policies]
                backends.append(bs_dict)
            ret["backends"] = backends
            ret["subnets"] = lb.subnets
            ret["security_groups"] = lb.security_groups
            ret["scheme"] = lb.scheme
            ret["dns_name"] = lb.dns_name
            ret["tags"] = _get_all_tags(conn, name)
            lb_policy_lists = [
                lb.policies.app_cookie_stickiness_policies,
                lb.policies.lb_cookie_stickiness_policies,
                lb.policies.other_policies,
            ]
            policies = []
            for policy_list in lb_policy_lists:
                policies += [p.policy_name for p in policy_list]
            ret["policies"] = policies
            ret["canonical_hosted_zone_name"] = lb.canonical_hosted_zone_name
            ret["canonical_hosted_zone_name_id"] = lb.canonical_hosted_zone_name_id
            ret["vpc_id"] = lb.vpc_id
            return ret
        except boto.exception.BotoServerError as error:
            if error.error_code == "Throttling":
                log.debug("Throttled by AWS API, will retry in 5 seconds.")
                time.sleep(5)
                retries -= 1
                continue
            log.error("Error fetching config for ELB %s: %s", name, error.message)
            log.error(error)
            return {}
    return {}


def listener_dict_to_tuple(listener):
    """
    Convert an ELB listener dict into a listener tuple used by certain parts of
    the AWS ELB API.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.listener_dict_to_tuple '{"elb_port":80,"instance_port":80,"elb_protocol":"HTTP"}'
    """
    # We define all listeners as complex listeners.
    if "instance_protocol" not in listener:
        instance_protocol = listener["elb_protocol"].upper()
    else:
        instance_protocol = listener["instance_protocol"].upper()
    listener_tuple = [
        listener["elb_port"],
        listener["instance_port"],
        listener["elb_protocol"],
        instance_protocol,
    ]
    if "certificate" in listener:
        listener_tuple.append(listener["certificate"])
    return tuple(listener_tuple)


def create(
    name,
    availability_zones,
    listeners,
    subnets=None,
    security_groups=None,
    scheme="internet-facing",
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an ELB

    CLI example to create an ELB:

    .. code-block:: bash

        salt myminion boto_elb.create myelb '["us-east-1a", "us-east-1e"]' '{"elb_port": 443, "elb_protocol": "HTTPS", ...}' region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if exists(name, region, key, keyid, profile):
        return True
    if isinstance(availability_zones, str):
        availability_zones = salt.utils.json.loads(availability_zones)

    if isinstance(listeners, str):
        listeners = salt.utils.json.loads(listeners)

    _complex_listeners = []
    for listener in listeners:
        _complex_listeners.append(listener_dict_to_tuple(listener))

    try:
        lb = conn.create_load_balancer(
            name=name,
            zones=availability_zones,
            subnets=subnets,
            security_groups=security_groups,
            scheme=scheme,
            complex_listeners=_complex_listeners,
        )
        if lb:
            log.info("Created ELB %s", name)
            return True
        else:
            log.error("Failed to create ELB %s", name)
            return False
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to create ELB %s: %s: %s",
            name,
            error.error_code,
            error.message,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def delete(name, region=None, key=None, keyid=None, profile=None):
    """
    Delete an ELB.

    CLI example to delete an ELB:

    .. code-block:: bash

        salt myminion boto_elb.delete myelb region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if not exists(name, region, key, keyid, profile):
        return True
    try:
        conn.delete_load_balancer(name)
        log.info("Deleted ELB %s.", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error("Failed to delete ELB %s", name, exc_info_on_loglevel=logging.DEBUG)
        return False


def create_listeners(name, listeners, region=None, key=None, keyid=None, profile=None):
    """
    Create listeners on an ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.create_listeners myelb '[["HTTPS", "HTTP", 443, 80, "arn:aws:iam::11  11111:server-certificate/mycert"]]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(listeners, str):
        listeners = salt.utils.json.loads(listeners)

    _complex_listeners = []
    for listener in listeners:
        _complex_listeners.append(listener_dict_to_tuple(listener))
    try:
        conn.create_load_balancer_listeners(name, [], _complex_listeners)
        log.info("Created ELB listeners on %s", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to create ELB listeners on %s: %s",
            name,
            error,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def delete_listeners(name, ports, region=None, key=None, keyid=None, profile=None):
    """
    Delete listeners on an ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.delete_listeners myelb '[80,443]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(ports, str):
        ports = salt.utils.json.loads(ports)
    try:
        conn.delete_load_balancer_listeners(name, ports)
        log.info("Deleted ELB listeners on %s", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to delete ELB listeners on %s: %s",
            name,
            error,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def apply_security_groups(
    name, security_groups, region=None, key=None, keyid=None, profile=None
):
    """
    Apply security groups to ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.apply_security_groups myelb '["mysecgroup1"]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(security_groups, str):
        security_groups = salt.utils.json.loads(security_groups)
    try:
        conn.apply_security_groups_to_lb(name, security_groups)
        log.info("Applied security_groups on ELB %s", name)
        return True
    except boto.exception.BotoServerError as e:
        log.debug(e)
        log.error("Failed to appply security_groups on ELB %s: %s", name, e.message)
        return False


def enable_availability_zones(
    name, availability_zones, region=None, key=None, keyid=None, profile=None
):
    """
    Enable availability zones for ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.enable_availability_zones myelb '["us-east-1a"]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(availability_zones, str):
        availability_zones = salt.utils.json.loads(availability_zones)
    try:
        conn.enable_availability_zones(name, availability_zones)
        log.info("Enabled availability_zones on ELB %s", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error("Failed to enable availability_zones on ELB %s: %s", name, error)
        return False


def disable_availability_zones(
    name, availability_zones, region=None, key=None, keyid=None, profile=None
):
    """
    Disable availability zones for ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.disable_availability_zones myelb '["us-east-1a"]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(availability_zones, str):
        availability_zones = salt.utils.json.loads(availability_zones)
    try:
        conn.disable_availability_zones(name, availability_zones)
        log.info("Disabled availability_zones on ELB %s", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to disable availability_zones on ELB %s: %s",
            name,
            error,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def attach_subnets(name, subnets, region=None, key=None, keyid=None, profile=None):
    """
    Attach ELB to subnets.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.attach_subnets myelb '["mysubnet"]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(subnets, str):
        subnets = salt.utils.json.loads(subnets)
    try:
        conn.attach_lb_to_subnets(name, subnets)
        log.info("Attached ELB %s on subnets.", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to attach ELB %s on subnets: %s",
            name,
            error,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def detach_subnets(name, subnets, region=None, key=None, keyid=None, profile=None):
    """
    Detach ELB from subnets.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.detach_subnets myelb '["mysubnet"]'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if isinstance(subnets, str):
        subnets = salt.utils.json.loads(subnets)
    try:
        conn.detach_lb_from_subnets(name, subnets)
        log.info("Detached ELB %s from subnets.", name)
        return True
    except boto.exception.BotoServerError as error:
        log.error(
            "Failed to detach ELB %s from subnets: %s",
            name,
            error,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def get_attributes(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if attributes are set on an ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.get_attributes myelb
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30

    while retries:
        try:
            lbattrs = conn.get_all_lb_attributes(name)
            ret = OrderedDict()
            ret["access_log"] = OrderedDict()
            ret["cross_zone_load_balancing"] = OrderedDict()
            ret["connection_draining"] = OrderedDict()
            ret["connecting_settings"] = OrderedDict()
            al = lbattrs.access_log
            czlb = lbattrs.cross_zone_load_balancing
            cd = lbattrs.connection_draining
            cs = lbattrs.connecting_settings
            ret["access_log"]["enabled"] = al.enabled
            ret["access_log"]["s3_bucket_name"] = al.s3_bucket_name
            ret["access_log"]["s3_bucket_prefix"] = al.s3_bucket_prefix
            ret["access_log"]["emit_interval"] = al.emit_interval
            ret["cross_zone_load_balancing"]["enabled"] = czlb.enabled
            ret["connection_draining"]["enabled"] = cd.enabled
            ret["connection_draining"]["timeout"] = cd.timeout
            ret["connecting_settings"]["idle_timeout"] = cs.idle_timeout
            return ret
        except boto.exception.BotoServerError as e:
            if e.error_code == "Throttling":
                log.debug("Throttled by AWS API, will retry in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error("ELB %s does not exist: %s", name, e.message)
            return {}
    return {}


def set_attributes(name, attributes, region=None, key=None, keyid=None, profile=None):
    """
    Set attributes on an ELB.

    name (string)
        Name of the ELB instance to set attributes for

    attributes
        A dict of attributes to set.

        Valid attributes are:

        access_log (dict)
            enabled (bool)
                Enable storage of access logs.
            s3_bucket_name (string)
                The name of the S3 bucket to place logs.
            s3_bucket_prefix (string)
                Prefix for the log file name.
            emit_interval (int)
                Interval for storing logs in S3 in minutes. Valid values are
                5 and 60.

        connection_draining (dict)
            enabled (bool)
                Enable connection draining.
            timeout (int)
                Maximum allowed time in seconds for sending existing
                connections to an instance that is deregistering or unhealthy.
                Default is 300.

        cross_zone_load_balancing (dict)
            enabled (bool)
                Enable cross-zone load balancing.

    CLI example to set attributes on an ELB:

    .. code-block:: bash

        salt myminion boto_elb.set_attributes myelb '{"access_log": {"enabled": "true", "s3_bucket_name": "mybucket", "s3_bucket_prefix": "mylogs/", "emit_interval": "5"}}' region=us-east-1
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    al = attributes.get("access_log", {})
    czlb = attributes.get("cross_zone_load_balancing", {})
    cd = attributes.get("connection_draining", {})
    cs = attributes.get("connecting_settings", {})
    if not al and not czlb and not cd and not cs:
        log.error("No supported attributes for ELB.")
        return False
    if al:
        _al = AccessLogAttribute()
        _al.enabled = al.get("enabled", False)
        if not _al.enabled:
            msg = "Access log attribute configured, but enabled config missing"
            log.error(msg)
            return False
        _al.s3_bucket_name = al.get("s3_bucket_name", None)
        _al.s3_bucket_prefix = al.get("s3_bucket_prefix", None)
        _al.emit_interval = al.get("emit_interval", None)
        added_attr = conn.modify_lb_attribute(name, "accessLog", _al)
        if added_attr:
            log.info("Added access_log attribute to %s elb.", name)
        else:
            log.error("Failed to add access_log attribute to %s elb.", name)
            return False
    if czlb:
        _czlb = CrossZoneLoadBalancingAttribute()
        _czlb.enabled = czlb["enabled"]
        added_attr = conn.modify_lb_attribute(
            name, "crossZoneLoadBalancing", _czlb.enabled
        )
        if added_attr:
            log.info("Added cross_zone_load_balancing attribute to %s elb.", name)
        else:
            log.error("Failed to add cross_zone_load_balancing attribute.")
            return False
    if cd:
        _cd = ConnectionDrainingAttribute()
        _cd.enabled = cd["enabled"]
        _cd.timeout = cd.get("timeout", 300)
        added_attr = conn.modify_lb_attribute(name, "connectionDraining", _cd)
        if added_attr:
            log.info("Added connection_draining attribute to %s elb.", name)
        else:
            log.error("Failed to add connection_draining attribute.")
            return False
    if cs:
        _cs = ConnectionSettingAttribute()
        _cs.idle_timeout = cs.get("idle_timeout", 60)
        added_attr = conn.modify_lb_attribute(name, "connectingSettings", _cs)
        if added_attr:
            log.info("Added connecting_settings attribute to %s elb.", name)
        else:
            log.error("Failed to add connecting_settings attribute.")
            return False
    return True


def get_health_check(name, region=None, key=None, keyid=None, profile=None):
    """
    Get the health check configured for this ELB.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.get_health_check myelb
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30

    while True:
        try:
            lb = conn.get_all_load_balancers(load_balancer_names=[name])
            lb = lb[0]
            ret = OrderedDict()
            hc = lb.health_check
            ret["interval"] = hc.interval
            ret["target"] = hc.target
            ret["healthy_threshold"] = hc.healthy_threshold
            ret["timeout"] = hc.timeout
            ret["unhealthy_threshold"] = hc.unhealthy_threshold
            return ret
        except boto.exception.BotoServerError as e:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, will retry in 5 seconds.")
                time.sleep(5)
                retries -= 1
                continue
            log.error("ELB %s not found.", name, exc_info_on_logleve=logging.DEBUG)
            return {}


def set_health_check(
    name, health_check, region=None, key=None, keyid=None, profile=None
):
    """
    Set attributes on an ELB.

    CLI example to set attributes on an ELB:

    .. code-block:: bash

        salt myminion boto_elb.set_health_check myelb '{"target": "HTTP:80/"}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    retries = 30

    hc = HealthCheck(**health_check)
    while True:
        try:
            conn.configure_health_check(name, hc)
            log.info("Configured health check on ELB %s", name)
            return True
        except boto.exception.BotoServerError as error:
            if retries and e.code == "Throttling":
                log.debug("Throttled by AWS API, will retry in 5 seconds.")
                time.sleep(5)
                retries -= 1
                continue
            log.exception("Failed to configure health check on ELB %s", name)
            return False


def register_instances(
    name, instances, region=None, key=None, keyid=None, profile=None
):
    """
    Register instances with an ELB.  Instances is either a string
    instance id or a list of string instance id's.

    Returns:

    - ``True``: instance(s) registered successfully
    - ``False``: instance(s) failed to be registered

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.register_instances myelb instance_id
        salt myminion boto_elb.register_instances myelb "[instance_id,instance_id]"
    """
    # convert instances to list type, enabling consistent use of instances
    # variable throughout the register_instances method
    if isinstance(instances, str) or isinstance(instances, str):
        instances = [instances]
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        registered_instances = conn.register_instances(name, instances)
    except boto.exception.BotoServerError as error:
        log.warning(error)
        return False
    registered_instance_ids = [instance.id for instance in registered_instances]
    # register_failues is a set that will contain any instances that were not
    # able to be registered with the given ELB
    register_failures = set(instances).difference(set(registered_instance_ids))
    if register_failures:
        log.warning(
            "Instance(s): %s not registered with ELB %s.", list(register_failures), name
        )
        register_result = False
    else:
        register_result = True
    return register_result


def deregister_instances(
    name, instances, region=None, key=None, keyid=None, profile=None
):
    """
    Deregister instances with an ELB.  Instances is either a string
    instance id or a list of string instance id's.

    Returns:

    - ``True``: instance(s) deregistered successfully
    - ``False``: instance(s) failed to be deregistered
    - ``None``: instance(s) not valid or not registered, no action taken

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.deregister_instances myelb instance_id
        salt myminion boto_elb.deregister_instances myelb "[instance_id, instance_id]"
    """
    # convert instances to list type, enabling consistent use of instances
    # variable throughout the deregister_instances method
    if isinstance(instances, str) or isinstance(instances, str):
        instances = [instances]
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        registered_instances = conn.deregister_instances(name, instances)
    except boto.exception.BotoServerError as error:
        # if the instance(s) given as an argument are not members of the ELB
        # boto returns error.error_code == 'InvalidInstance'
        # deregister_instances returns "None" because the instances are
        # effectively deregistered from ELB
        if error.error_code == "InvalidInstance":
            log.warning(
                "One or more of instance(s) %s are not part of ELB %s. "
                "deregister_instances not performed.",
                instances,
                name,
            )
            return None
        else:
            log.warning(error)
            return False
    registered_instance_ids = [instance.id for instance in registered_instances]
    # deregister_failures is a set that will contain any instances that were
    # unable to be deregistered from the given ELB
    deregister_failures = set(instances).intersection(set(registered_instance_ids))
    if deregister_failures:
        log.warning(
            "Instance(s): %s not deregistered from ELB %s.",
            list(deregister_failures),
            name,
        )
        deregister_result = False
    else:
        deregister_result = True
    return deregister_result


def set_instances(
    name, instances, test=False, region=None, key=None, keyid=None, profile=None
):
    """
    Set the instances assigned to an ELB to exactly the list given

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.set_instances myelb region=us-east-1 instances="[instance_id,instance_id]"
    """
    ret = True
    current = {
        i["instance_id"] for i in get_instance_health(name, region, key, keyid, profile)
    }
    desired = set(instances)
    add = desired - current
    remove = current - desired
    if test:
        return bool(add or remove)
    if remove:
        if (
            deregister_instances(name, list(remove), region, key, keyid, profile)
            is False
        ):
            ret = False
    if add:
        if register_instances(name, list(add), region, key, keyid, profile) is False:
            ret = False
    return ret


def get_instance_health(
    name, region=None, key=None, keyid=None, profile=None, instances=None
):
    """
    Get a list of instances and their health state

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.get_instance_health myelb
        salt myminion boto_elb.get_instance_health myelb region=us-east-1 instances="[instance_id,instance_id]"
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        instance_states = conn.describe_instance_health(name, instances)
        ret = []
        for _instance in instance_states:
            ret.append(
                {
                    "instance_id": _instance.instance_id,
                    "description": _instance.description,
                    "state": _instance.state,
                    "reason_code": _instance.reason_code,
                }
            )
        return ret
    except boto.exception.BotoServerError as error:
        log.debug(error)
        return []


def create_policy(
    name,
    policy_name,
    policy_type,
    policy,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an ELB policy.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.create_policy myelb mypolicy LBCookieStickinessPolicyType '{"CookieExpirationPeriod": 3600}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if not exists(name, region, key, keyid, profile):
        return False
    try:
        success = conn.create_lb_policy(name, policy_name, policy_type, policy)
        if success:
            log.info("Created policy %s on ELB %s", policy_name, name)
            return True
        else:
            log.error("Failed to create policy %s on ELB %s", policy_name, name)
            return False
    except boto.exception.BotoServerError as e:
        log.error(
            "Failed to create policy %s on ELB %s: %s",
            policy_name,
            name,
            e.message,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def delete_policy(name, policy_name, region=None, key=None, keyid=None, profile=None):
    """
    Delete an ELB policy.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.delete_policy myelb mypolicy
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if not exists(name, region, key, keyid, profile):
        return True
    try:
        conn.delete_lb_policy(name, policy_name)
        log.info("Deleted policy %s on ELB %s", policy_name, name)
        return True
    except boto.exception.BotoServerError as e:
        log.error(
            "Failed to delete policy %s on ELB %s: %s",
            policy_name,
            name,
            e.message,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False


def set_listener_policy(
    name, port, policies=None, region=None, key=None, keyid=None, profile=None
):
    """
    Set the policies of an ELB listener.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: Bash

        salt myminion boto_elb.set_listener_policy myelb 443 "[policy1,policy2]"
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if not exists(name, region, key, keyid, profile):
        return True
    if policies is None:
        policies = []
    try:
        conn.set_lb_policies_of_listener(name, port, policies)
        log.info("Set policies %s on ELB %s listener %s", policies, name, port)
    except boto.exception.BotoServerError as e:
        log.info(
            "Failed to set policy %s on ELB %s listener %s: %s",
            policies,
            name,
            port,
            e.message,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False
    return True


def set_backend_policy(
    name, port, policies=None, region=None, key=None, keyid=None, profile=None
):
    """
    Set the policies of an ELB backend server.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.set_backend_policy myelb 443 "[policy1,policy2]"
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if not exists(name, region, key, keyid, profile):
        return True
    if policies is None:
        policies = []
    try:
        conn.set_lb_policies_of_backend_server(name, port, policies)
        log.info("Set policies %s on ELB %s backend server %s", policies, name, port)
    except boto.exception.BotoServerError as e:
        log.info(
            "Failed to set policy %s on ELB %s backend server %s: %s",
            policies,
            name,
            port,
            e.message,
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False
    return True


def set_tags(name, tags, region=None, key=None, keyid=None, profile=None):
    """
    Add the tags on an ELB

    .. versionadded:: 2016.3.0

    name
        name of the ELB

    tags
        dict of name/value pair tags

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.set_tags my-elb-name "{'Tag1': 'Value', 'Tag2': 'Another Value'}"
    """

    if exists(name, region, key, keyid, profile):
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        ret = _add_tags(conn, name, tags)
        return ret
    else:
        return False


def delete_tags(name, tags, region=None, key=None, keyid=None, profile=None):
    """
    Add the tags on an ELB

    name
        name of the ELB

    tags
        list of tags to remove

    CLI Example:

    .. code-block:: bash

        salt myminion boto_elb.delete_tags my-elb-name ['TagToRemove1', 'TagToRemove2']
    """
    if exists(name, region, key, keyid, profile):
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
        ret = _remove_tags(conn, name, tags)
        return ret
    else:
        return False


def _build_tag_param_list(params, tags):
    """
    helper function to build a tag parameter list to send
    """
    keys = sorted(tags.keys())
    i = 1
    for key in keys:
        value = tags[key]
        params[f"Tags.member.{i}.Key"] = key
        if value is not None:
            params[f"Tags.member.{i}.Value"] = value
        i += 1


def _get_all_tags(conn, load_balancer_names=None):
    """
    Retrieve all the metadata tags associated with your ELB(s).

    :type load_balancer_names: list
    :param load_balancer_names: An optional list of load balancer names.

    :rtype: list
    :return: A list of :class:`boto.ec2.elb.tag.Tag` objects
    """
    params = {}
    if load_balancer_names:
        conn.build_list_params(
            params, load_balancer_names, "LoadBalancerNames.member.%d"
        )

    tags = conn.get_object(
        "DescribeTags",
        params,
        __utils__["boto_elb_tag.get_tag_descriptions"](),
        verb="POST",
    )
    if tags[load_balancer_names]:
        return tags[load_balancer_names]
    else:
        return None


def _add_tags(conn, load_balancer_names, tags):
    """
    Create new metadata tags for the specified resource ids.

    :type load_balancer_names: list
    :param load_balancer_names: A list of load balancer names.

    :type tags: dict
    :param tags: A dictionary containing the name/value pairs.
                 If you want to create only a tag name, the
                 value for that tag should be the empty string
                 (e.g. '').
    """
    params = {}
    conn.build_list_params(params, load_balancer_names, "LoadBalancerNames.member.%d")
    _build_tag_param_list(params, tags)
    return conn.get_status("AddTags", params, verb="POST")


def _remove_tags(conn, load_balancer_names, tags):
    """
    Delete metadata tags for the specified resource ids.

    :type load_balancer_names: list
    :param load_balancer_names: A list of load balancer names.

    :type tags: list
    :param tags: A list containing just tag names for the tags to be
                 deleted.
    """
    params = {}
    conn.build_list_params(params, load_balancer_names, "LoadBalancerNames.member.%d")
    conn.build_list_params(params, tags, "Tags.member.%d.Key")
    return conn.get_status("RemoveTags", params, verb="POST")



import logging
import os

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.plist as plistserializer
import salt.serializers.python as pythonserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "python.serialize": pythonserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "plist.serialize": plistserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


# 'comment' function tests: 1
def test_comment():
    """
    Test to comment out specified lines in a file.
    """
    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        name = "/etc/aliases" if salt.utils.platform.is_darwin() else "/etc/fstab"
        regex = "bind 127.0.0.1"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        comt = "Must provide name to file.comment"
        ret.update({"comment": comt, "name": ""})
        assert filestate.comment("", regex) == ret

        mock_t = MagicMock(return_value=True)
        mock_f = MagicMock(return_value=False)
        with patch.object(os.path, "isabs", mock_f):
            comt = f"Specified file {name} is not an absolute path"
            ret.update({"comment": comt, "name": name})
            assert filestate.comment(name, regex) == ret

        with patch.object(os.path, "isabs", mock_t):
            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": MagicMock(
                        side_effect=[False, True, False, False, False, False]
                    )
                },
            ):
                comt = "Pattern already commented"
                ret.update({"comment": comt, "result": True})
                assert filestate.comment(name, regex) == ret

                comt = "Pattern not found and ignore_missing set to True"
                ret.update({"comment": comt, "result": True})
                assert filestate.comment(name, regex, ignore_missing=True) == ret

                comt = f"{regex}: Pattern not found"
                ret.update({"comment": comt, "result": False})
                assert filestate.comment(name, regex) == ret

            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": MagicMock(
                        side_effect=[True, True, True, False, True]
                    ),
                    "file.comment": mock_t,
                    "file.comment_line": mock_t,
                },
            ):
                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = f"File {name} is set to be updated"
                    ret.update(
                        {"comment": comt, "result": None, "changes": {name: "updated"}}
                    )
                    assert filestate.comment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Commented lines successfully"
                        ret.update({"comment": comt, "result": True, "changes": {}})
                        assert filestate.comment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = "Pattern already commented"
                    ret.update({"comment": comt, "result": True, "changes": {}})
                    assert filestate.comment(name, regex) == ret


# 'uncomment' function tests: 1
def test_uncomment():
    """
    Test to uncomment specified commented lines in a file
    """
    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        name = "/etc/aliases" if salt.utils.platform.is_darwin() else "/etc/fstab"
        regex = "bind 127.0.0.1"

        ret = {"name": name, "result": False, "comment": "", "changes": {}}

        comt = "Must provide name to file.uncomment"
        ret.update({"comment": comt, "name": ""})
        assert filestate.uncomment("", regex) == ret

        mock_t = MagicMock(return_value=True)
        mock_f = MagicMock(return_value=False)
        mock = MagicMock(side_effect=[False, True, False, False, True, True, True])
        with patch.object(os.path, "isabs", mock_f):
            comt = f"Specified file {name} is not an absolute path"
            ret.update({"comment": comt, "name": name})
            assert filestate.uncomment(name, regex) == ret

        with patch.object(os.path, "isabs", mock_t):
            with patch.dict(
                filestate.__salt__,
                {
                    "file.search": mock,
                    "file.uncomment": mock_t,
                    "file.comment_line": mock_t,
                },
            ):
                comt = "Pattern already uncommented"
                ret.update({"comment": comt, "result": True})
                assert filestate.uncomment(name, regex) == ret

                comt = f"{regex}: Pattern not found"
                ret.update({"comment": comt, "result": False})
                assert filestate.uncomment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": True}):
                    comt = f"File {name} is set to be updated"
                    ret.update(
                        {"comment": comt, "result": None, "changes": {name: "updated"}}
                    )
                    assert filestate.uncomment(name, regex) == ret

                with patch.dict(filestate.__opts__, {"test": False}):
                    with patch.object(
                        salt.utils.files, "fopen", MagicMock(mock_open())
                    ):
                        comt = "Uncommented lines successfully"
                        ret.update({"comment": comt, "result": True, "changes": {}})
                        assert filestate.uncomment(name, regex) == ret



import logging
import os

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.plist as plistserializer
import salt.serializers.python as pythonserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "python.serialize": pythonserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "plist.serialize": plistserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


def test_symlink():
    """
    Test to create a symlink.
    """
    name = os.sep + os.path.join("tmp", "testfile.txt")
    target = salt.utils.files.mkstemp()
    test_dir = os.sep + "tmp"
    user = "salt"

    if salt.utils.platform.is_windows():
        group = "salt"
    else:
        group = "saltstack"

    def return_val(kwargs):
        val = {
            "name": name,
            "result": False,
            "comment": "",
            "changes": {},
        }
        val.update(kwargs)
        return val

    mock_t = MagicMock(return_value=True)
    mock_f = MagicMock(return_value=False)
    mock_empty = MagicMock(return_value="")
    mock_uid = MagicMock(return_value="U1001")
    mock_gid = MagicMock(return_value="g1001")
    mock_target = MagicMock(return_value=target)
    mock_user = MagicMock(return_value=user)
    mock_grp = MagicMock(return_value=group)
    mock_os_error = MagicMock(side_effect=OSError)

    with patch.dict(filestate.__salt__, {"config.manage_mode": mock_t}):
        comt = "Must provide name to file.symlink"
        ret = return_val({"comment": comt, "name": ""})
        assert filestate.symlink("", target) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_empty,
            "file.group_to_gid": mock_empty,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ):
        if salt.utils.platform.is_windows():
            comt = f"User {user} does not exist"
            ret = return_val({"comment": comt, "name": name})
        else:
            comt = "User {} does not exist. Group {} does not exist.".format(
                user, group
            )
            ret = return_val({"comment": comt, "name": name})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": True}), patch.object(
        os.path, "exists", mock_f
    ):
        if salt.utils.platform.is_windows():
            comt = f"User {user} does not exist"
            ret = return_val(
                {"comment": comt, "result": False, "name": name, "changes": {}}
            )
        else:
            comt = f"Symlink {name} to {target} is set for creation"
            ret = return_val(
                {"comment": comt, "result": None, "changes": {"new": name}}
            )
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", mock_f
    ), patch.object(
        os.path, "exists", mock_f
    ):
        if salt.utils.platform.is_windows():
            comt = f"User {user} does not exist"
            ret = return_val(
                {"comment": comt, "result": False, "name": name, "changes": {}}
            )
        else:
            comt = f"Directory {test_dir} for symlink is not present"
            ret = return_val({"comment": comt, "result": False, "changes": {}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_t,
            "file.readlink": mock_target,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", mock_t
    ), patch.object(
        salt.states.file, "_check_symlink_ownership", mock_t
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        if salt.utils.platform.is_windows():
            comt = f"Symlink {name} is present and owned by {user}"
        else:
            comt = f"Symlink {name} is present and owned by {user}:{group}"
        ret = return_val({"comment": comt, "result": True, "changes": {}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", mock_t
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.object(
        os.path, "lexists", mock_t
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = (
            "Symlink & backup dest exists and Force not set. {} -> "
            "{} - backup: {}".format(name, target, os.path.join(test_dir, "SALT"))
        )
        ret.update({"comment": comt, "result": False, "changes": {}})
        assert (
            filestate.symlink(name, target, user=user, group=group, backupname="SALT")
            == ret
        )

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "exists", mock_t
    ), patch.object(
        os.path, "isfile", mock_t
    ), patch.object(
        os.path, "isdir", mock_t
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = "Backupname must be an absolute path or a file name: {}".format(
            "tmp/SALT"
        )
        ret.update({"comment": comt, "result": False, "changes": {}})
        assert (
            filestate.symlink(
                name, target, user=user, group=group, backupname="tmp/SALT"
            )
            == ret
        )

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "user.info": mock_empty,
            "user.current": mock_user,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", mock_t
    ), patch.object(
        os.path, "exists", mock_t
    ), patch.object(
        os.path, "isfile", mock_t
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = f"File exists where the symlink {name} should be"
        ret = return_val({"comment": comt, "changes": {}, "result": False})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "file.symlink": mock_t,
            "user.info": mock_t,
            "file.lchown": mock_f,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isdir", mock_t
    ), patch.object(
        os.path, "exists", mock_t
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = f"Directory exists where the symlink {name} should be"
        ret = return_val({"comment": comt, "result": False, "changes": {}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "file.symlink": mock_os_error,
            "user.info": mock_t,
            "file.lchown": mock_f,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isfile", mock_f
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = f"Unable to create new symlink {name} -> {target}: "
        ret = return_val({"comment": comt, "result": False, "changes": {}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "file.symlink": mock_t,
            "user.info": mock_t,
            "file.lchown": mock_f,
            "file.get_user": mock_user,
            "file.get_group": mock_grp,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isfile", mock_f
    ), patch(
        "salt.states.file._check_symlink_ownership", return_value=True
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ):
        comt = f"Created new symlink {name} -> {target}"
        ret = return_val({"comment": comt, "result": True, "changes": {"new": name}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "file.symlink": mock_t,
            "user.info": mock_t,
            "file.lchown": mock_f,
            "file.get_user": mock_empty,
            "file.get_group": mock_empty,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isfile", mock_f
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ), patch(
        "salt.states.file._set_symlink_ownership", return_value=False
    ), patch(
        "salt.states.file._check_symlink_ownership", return_value=False
    ):
        comt = (
            "Created new symlink {} -> {}, but was unable to set "
            "ownership to {}:{}".format(name, target, user, group)
        )
        ret = return_val({"comment": comt, "result": False, "changes": {"new": name}})
        assert filestate.symlink(name, target, user=user, group=group) == ret

    with patch.dict(
        filestate.__salt__,
        {
            "config.manage_mode": mock_t,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.is_link": mock_f,
            "file.readlink": mock_target,
            "file.symlink": mock_t,
            "user.info": mock_t,
            "file.lchown": mock_f,
            "file.get_user": mock_empty,
            "file.get_group": mock_empty,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isfile", mock_f
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ), patch(
        "salt.states.file._set_symlink_ownership", return_value=True
    ), patch(
        "salt.states.file._check_symlink_ownership", return_value=True
    ):
        comt = f"Created new symlink {name} -> {target}"
        ret = return_val({"comment": comt, "result": True, "changes": {"new": name}})
        res = filestate.symlink(name, target, user=user, group=user)
        assert res == ret

    with patch.dict(
        filestate.__salt__,
        {
            "file.is_link": mock_t,
            "file.get_user": mock_user,
            "file.get_group": mock_grp,
            "file.user_to_uid": mock_uid,
            "file.group_to_gid": mock_gid,
            "file.gid_to_group": MagicMock(return_value=group),
            "file.readlink": mock_target,
            "user.info": mock_t,
        },
    ), patch.dict(filestate.__opts__, {"test": False}), patch.object(
        os.path, "isdir", MagicMock(side_effect=[True, False])
    ), patch.object(
        os.path, "isfile", mock_f
    ), patch(
        "salt.utils.win_functions.get_sid_from_name", return_value="test-sid"
    ), patch(
        "salt.states.file._set_symlink_ownership", return_value=True
    ), patch(
        "salt.states.file._check_symlink_ownership", return_value=True
    ), patch(
        "salt.states.file._get_symlink_ownership", return_value=(user, group)
    ):
        if salt.utils.platform.is_windows():
            comt = f"Symlink {name} is present and owned by {user}"
        else:
            comt = f"Symlink {name} is present and owned by {user}:{group}"
        ret = return_val({"comment": comt, "result": True, "changes": {}})
        res = filestate.symlink(name, target, inherit_user_and_group=True)
        assert res == ret



import filecmp

import pytest

import salt.utils.files

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_issue_25250_force_copy_deletes(file, tmp_path):
    """
    ensure force option in copy state does not delete target file
    """
    source = tmp_path / "source"
    dest = tmp_path / "dest"
    source.write_text("source")
    dest.write_text("destination")

    ret = file.copy(name=str(dest), source=str(source), force=True)
    assert ret.result is True
    assert filecmp.cmp(str(source), str(dest)) is True


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows(reason="Windows does not report any file modes.")
def test_file_copy_make_dirs(modules, file, tmp_path, state_file_account):
    """
    ensure make_dirs creates correct user perms
    """
    source = tmp_path / "source"
    source.write_text("source")
    dest = tmp_path / "dir1" / "dir2" / "dest"

    mode = "0644"
    ret = file.copy(
        name=str(dest),
        source=str(source),
        user=state_file_account.username,
        makedirs=True,
        mode=mode,
    )
    assert ret.result is True
    file_checks = [
        (str(dest), mode),
        (str(dest.parent), "0755"),
        (str(dest.parent.parent), "0755"),
    ]
    for check, expected_mode in file_checks:
        user_check = modules.file.get_user(check)
        mode_check = modules.file.get_mode(check)
        assert user_check == state_file_account.username
        assert salt.utils.files.normalize_mode(mode_check) == expected_mode



"""
sysrc module for FreeBSD
"""

import salt.utils.path
from salt.exceptions import CommandExecutionError

__virtualname__ = "sysrc"

__func_alias__ = {"set_": "set"}


def __virtual__():
    """
    Only runs if sysrc exists
    """
    if salt.utils.path.which("sysrc") is not None:
        return True
    return (
        False,
        "The sysrc execution module failed to load: the sysrc binary is not in the"
        " path.",
    )


def get(**kwargs):
    """
    Return system rc configuration variables

    CLI Example:

     .. code-block:: bash

         salt '*' sysrc.get includeDefaults=True
    """

    cmd = "sysrc -v"

    if "file" in kwargs:
        cmd += " -f " + kwargs["file"]

    if "jail" in kwargs:
        cmd += " -j " + kwargs["jail"]

    if "name" in kwargs:
        cmd += " " + kwargs["name"]
    elif kwargs.get("includeDefaults", False):
        cmd += " -A"
    else:
        cmd += " -a"

    sysrcs = __salt__["cmd.run"](cmd)
    if "sysrc: unknown variable" in sysrcs:
        # raise CommandExecutionError(sysrcs)
        return None

    ret = {}
    for sysrc in sysrcs.split("\n"):
        line_components = sysrc.split(": ")
        rcfile = line_components[0]
        if len(line_components) > 2:
            var = line_components[1]
            val = line_components[2]
        else:
            var = line_components[1].rstrip(":")
            val = ""
        if rcfile not in ret:
            ret[rcfile] = {}
        ret[rcfile][var] = val
    return ret


def set_(name, value, **kwargs):
    """
    Set system rc configuration variables

    CLI Example:

     .. code-block:: bash

         salt '*' sysrc.set name=sshd_flags value="-p 2222"
    """

    cmd = "sysrc -v"

    if "file" in kwargs:
        cmd += " -f " + kwargs["file"]

    if "jail" in kwargs:
        cmd += " -j " + kwargs["jail"]

    # This is here because the YAML parser likes to convert the string literals
    # YES, NO, Yes, No, True, False, etc. to boolean types.  However, in this case,
    # we will check to see if that happened and replace it with "YES" or "NO" because
    # those items are accepted in sysrc.
    if isinstance(value, bool):
        if value:
            value = "YES"
        else:
            value = "NO"

    # This is here for the same reason, except for numbers
    if isinstance(value, int):
        value = str(value)

    cmd += " " + name + '="' + value + '"'

    r = __salt__["cmd.run_all"](cmd)

    if r["retcode"] != 0:
        raise CommandExecutionError("sysrc failed: {}".format(r["stderr"]))

    ret = {}
    for sysrc in r["stdout"].split("\n"):
        rcfile = sysrc.split(": ")[0]
        var = sysrc.split(": ")[1]
        oldval = sysrc.split(": ")[2].strip().split("->")[0]
        newval = sysrc.split(": ")[2].strip().split("->")[1]
        if rcfile not in ret:
            ret[rcfile] = {}
        ret[rcfile][var] = newval
    return ret


def remove(name, **kwargs):
    """
    Remove system rc configuration variables

    CLI Example:

     .. code-block:: bash

         salt '*' sysrc.remove name=sshd_enable
    """

    cmd = "sysrc -v"

    if "file" in kwargs:
        cmd += " -f " + kwargs["file"]

    if "jail" in kwargs:
        cmd += " -j " + kwargs["jail"]

    cmd += " -x " + name

    sysrcs = __salt__["cmd.run"](cmd)
    if "sysrc: unknown variable" in sysrcs:
        raise CommandExecutionError(sysrcs)
    else:
        return name + " removed"



import ctypes
import os
import platform

import pytest

import salt.modules.win_useradd
import salt.utils.files
from salt.exceptions import CommandExecutionError
from salt.modules import win_file

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules(minion_opts, modules):
    utils = salt.loader.utils(minion_opts)
    return {
        win_file: {
            "__salt__": modules,
            "__utils__": utils,
        },
    }


skip_not_windows_admin = pytest.mark.skipif(
    platform.system() != "Windows" or not ctypes.windll.shell32.IsUserAnAdmin(),
    reason="user needs windows admin rights",
)


@pytest.fixture
def windows_user():
    return salt.modules.win_useradd.current()


@skip_not_windows_admin
def test_symlink(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    link = tmp_path / "l"
    assert win_file.is_link(str(link)) is False
    assert win_file.symlink(str(file), str(link)) is True
    assert file.is_file() is True
    assert win_file.is_link(str(link)) is True


@skip_not_windows_admin
def test_symlink_path_taken(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    link = tmp_path / "l"
    link.touch()
    assert link.is_file() is True
    # symlink should raise error if path name is all ready taken
    pytest.raises(CommandExecutionError, win_file.symlink, str(file), str(link))


@skip_not_windows_admin
def test_symlink_force(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    link = tmp_path / "l"
    assert win_file.is_link(str(link)) is False
    assert win_file.symlink(str(file), str(link), force=True) is True
    assert file.is_file() is True
    assert win_file.is_link(str(link)) is True
    # check that symlink returns ture if link is all ready-made
    assert win_file.symlink(str(file), str(link), force=True) is True


@skip_not_windows_admin
def test_symlink_atomic(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    link = tmp_path / "l"
    assert win_file.is_link(str(link)) is False
    assert win_file.symlink(str(file), str(link), force=True, atomic=True) is True
    assert file.is_file() is True
    assert win_file.is_link(str(link)) is True
    # check that symlink returns ture if link is all ready-made
    assert win_file.symlink(str(file), str(link), force=True, atomic=True) is True


def test_is_not_link(tmp_path):
    tmp_path = str(tmp_path)
    assert win_file.is_link(tmp_path) is False
    assert win_file.is_link(os.path.join(tmp_path, "made_up_path")) is False


@skip_not_windows_admin
def test__resolve_symlink(tmp_path):
    link = tmp_path / "l"
    assert win_file.symlink(str(tmp_path), str(link)) is True
    assert win_file._resolve_symlink(str(link)) == str(tmp_path)


def test_get_user(tmp_path, windows_user):
    path = tmp_path / "dir"
    assert win_file.mkdir(str(path), owner=windows_user) is True
    assert path.is_dir() is True
    assert win_file.get_user(str(path)) in windows_user


def test_fake_user(tmp_path, windows_user):
    path = tmp_path / "dir"
    windows_user = windows_user + "_fake"
    pytest.raises(
        CommandExecutionError,
        win_file.mkdir,
        str(path),
        owner=windows_user,
    )


def test_uid_user(tmp_path):
    path = str(tmp_path)
    uid = win_file.get_uid(path)
    assert "-" in uid
    assert win_file.user_to_uid(win_file.uid_to_user(uid)) == uid


def test_uid_to_user_none(tmp_path):
    assert win_file.uid_to_user(None) == ""


def test_gid_group(tmp_path, windows_user):
    path = str(tmp_path)
    gid = win_file.get_gid(path)
    assert isinstance(gid, str)
    group = win_file.get_group(path)
    assert isinstance(group, str)
    assert win_file.gid_to_group(gid) == group
    assert win_file.group_to_gid(group) == gid


def test_gid_group_none(tmp_path):
    assert win_file.group_to_gid(None) == ""


def test_get_pgroup(tmp_path):
    path = str(tmp_path)
    pgroup = win_file.get_pgroup(path)
    assert pgroup == "None"


def test_get_pgid(tmp_path):
    path = str(tmp_path)
    pgid = win_file.get_pgid(path)
    assert "-" in pgid


def test_get_uid_path_not_found(tmp_path):
    path = tmp_path / "dir"
    pytest.raises(CommandExecutionError, win_file.get_uid, str(path))


def test_get_user_path_not_found(tmp_path):
    path = tmp_path / "dir"
    pytest.raises(CommandExecutionError, win_file.get_user, str(path))


def test_mode(tmp_path):
    assert win_file.get_mode(str(tmp_path)) is None


def test_mode_path_not_found(tmp_path):
    path = tmp_path / "dir"
    pytest.raises(CommandExecutionError, win_file.get_mode, str(path))


def test_lchown(tmp_path, windows_user):
    assert win_file.lchown(str(tmp_path), windows_user) is True
    assert win_file.get_user(str(tmp_path)) == windows_user


def test_chown(tmp_path, windows_user):
    assert win_file.chown(str(tmp_path), windows_user) is True
    assert win_file.get_user(str(tmp_path)) == windows_user


def test_chpgrp(tmp_path):
    assert win_file.chpgrp(str(tmp_path), "Administrators") is True
    assert win_file.get_pgroup(str(tmp_path)) == "Administrators"


def test_chgrp(tmp_path):
    assert win_file.chgrp(str(tmp_path), "Administrators") is None


def test_stats(tmp_path, windows_user):
    stats = win_file.stats(str(tmp_path))
    assert isinstance(stats, dict)
    assert stats["type"] == "dir"
    assert isinstance(stats["inode"], int)
    assert isinstance(stats["uid"], str)
    assert isinstance(stats["gid"], str)
    assert isinstance(stats["user"], str)
    assert isinstance(stats["group"], str)
    assert isinstance(stats["pgroup"], str)
    assert isinstance(stats["atime"], (float, int))
    assert isinstance(stats["mtime"], (float, int))
    assert isinstance(stats["ctime"], (float, int))
    assert isinstance(stats["size"], int)
    assert isinstance(stats["mode"], str)
    assert isinstance(stats["target"], str)


def test_stats_path_not_found(tmp_path, windows_user):
    path = tmp_path / "dir"
    pytest.raises(CommandExecutionError, win_file.stats, str(path))


def test_version():
    assert len(win_file.version("C:\\Windows\\System32\\cmd.exe").split(".")) == 4


def test_version_empty(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    assert win_file.version(str(file)) == ""


def test_version_path_not_found(tmp_path):
    file = tmp_path / "t.txt"
    assert file.is_file() is False
    pytest.raises(CommandExecutionError, win_file.version, str(file))


def test_version_dir(tmp_path):
    path = tmp_path / "dir"
    path.mkdir()
    pytest.raises(CommandExecutionError, win_file.version, str(path))


def test_version_details():
    details = win_file.version_details("C:\\Windows\\System32\\cmd.exe")
    assert isinstance(details, dict) is True
    assert details["Comments"] is None
    assert details["CompanyName"] == "Microsoft Corporation"
    assert isinstance(details["FileDescription"], str) is True


def test_get_attributes(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    attributes = win_file.get_attributes(str(file))
    assert isinstance(attributes, dict) is True
    assert attributes["archive"] is True
    assert attributes["reparsePoint"] is False
    assert attributes["compressed"] is False
    assert attributes["directory"] is False
    assert attributes["encrypted"] is False
    assert attributes["hidden"] is False
    assert attributes["normal"] is False
    assert attributes["notIndexed"] is False
    assert attributes["offline"] is False
    assert attributes["readonly"] is False
    assert attributes["system"] is False
    assert attributes["temporary"] is False
    assert attributes["mountedVolume"] is False
    assert attributes["symbolicLink"] is False


def test_set_attributes(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    assert (
        win_file.set_attributes(
            str(file),
            archive=True,
            hidden=True,
            normal=False,
            notIndexed=True,
            readonly=True,
            system=True,
            temporary=True,
        )
        is True
    )
    attributes = win_file.get_attributes(str(file))
    assert attributes["archive"] is True
    assert attributes["hidden"] is True
    assert attributes["normal"] is False
    assert attributes["notIndexed"] is True
    assert attributes["readonly"] is True
    assert attributes["system"] is True
    assert attributes["temporary"] is True
    assert (
        win_file.set_attributes(
            str(file),
            archive=False,
            hidden=False,
            normal=True,
            notIndexed=False,
            readonly=False,
            system=False,
            temporary=False,
        )
        is True
    )
    attributes = win_file.get_attributes(str(file))
    assert attributes["archive"] is False
    assert attributes["hidden"] is False
    assert attributes["normal"] is True
    assert attributes["notIndexed"] is False
    assert attributes["readonly"] is False
    assert attributes["system"] is False
    assert attributes["temporary"] is False


def test_set_mode(tmp_path):
    assert win_file.set_mode(str(tmp_path), "") is None


def test_remove(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    assert win_file.remove(str(file)) is True
    assert file.is_file() is False


def test_remove_force(tmp_path):
    file = tmp_path / "t.txt"
    file.touch()
    assert file.is_file() is True
    assert win_file.remove(str(file), force=True) is True
    assert file.is_file() is False


def test_mkdir(tmp_path):
    path = tmp_path / "dir"
    assert path.is_dir() is False
    assert win_file.mkdir(str(path)) is True
    assert path.is_dir() is True


def test_mkdir_error(tmp_path):
    # dirs can't contain illegal characters on Windows
    illegal_name = "illegal*name"
    path = tmp_path / illegal_name
    pytest.raises(CommandExecutionError, win_file.mkdir, str(path))
    assert path.is_dir() is False

    # cant make dir if parent is not made
    path = tmp_path / "a" / "b" / "c" / "salt"
    pytest.raises(CommandExecutionError, win_file.mkdir, str(path))
    assert path.is_dir() is False


def test_makedirs_(tmp_path):
    parent = tmp_path / "dir1" / "dir2"
    path = parent / "dir3"
    assert win_file.makedirs_(str(path)) is True
    assert parent.is_dir() is True
    assert path.is_dir() is False


def test_makedirs__path_exists(tmp_path):
    parent = tmp_path / "dir1" / "dir2"
    path = parent / "dir3"
    parent.mkdir(parents=True)
    assert parent.is_dir() is True
    # makedirs_ should return message that path already exists
    result = win_file.makedirs_(str(path))
    assert "already exists" in result


def test_makedirs_perms(tmp_path):
    path = tmp_path / "dir1" / "dir2"
    assert win_file.makedirs_perms(str(path)) is True
    assert path.is_dir() is True
    # make sure makedirs does not fail if path exists
    assert win_file.makedirs_perms(str(path)) is True
    assert path.is_dir() is True


def test_check_perms_path_not_found(tmp_path, windows_user):
    path = tmp_path / "dir1"
    assert path.is_dir() is False
    # check_perms will fail due to path not being made
    pytest.raises(
        CommandExecutionError, win_file.check_perms, str(path), {}, windows_user
    )


def test_set_perms(tmp_path, windows_user):
    perms = {windows_user: {"perms": "read_execute", "applies_to": "this_folder_only"}}
    expected = {
        "grant": {
            windows_user: {"applies_to": "this_folder_only", "perms": "read_execute"}
        }
    }
    result = win_file.set_perms(str(tmp_path), grant_perms=perms)
    assert result == expected



import socket

import pytest

import salt.modules.cmdmod
import salt.modules.win_file
import salt.modules.win_lgpo as win_lgpo
from salt.exceptions import CommandExecutionError
from tests.support.mock import patch

try:
    import win32security as ws

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
    pytest.mark.skipif(not HAS_WIN32, reason="Failed to import win32security"),
]


@pytest.fixture
def configure_loader_modules():
    return {
        win_lgpo: {
            "__salt__": {
                "cmd.run": salt.modules.cmdmod.run,
                "file.file_exists": salt.modules.win_file.file_exists,
                "file.remove": salt.modules.win_file.remove,
            },
        },
    }


@pytest.fixture(scope="module")
def pol_info():
    return win_lgpo._policy_info()


@pytest.mark.parametrize(
    "val, expected",
    (
        (0, False),
        (1, True),
        ("", False),
        ("text", True),
        ([], False),
        ([1, 2, 3], True),
    ),
)
def test_notEmpty(pol_info, val, expected):
    assert pol_info._notEmpty(val) is expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        (0, 0),
        (86400, 1),
    ),
)
def test_seconds_to_days(pol_info, val, expected):
    assert pol_info._seconds_to_days(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        (0, 0),
        (1, 86400),
    ),
)
def test_days_to_seconds(pol_info, val, expected):
    assert pol_info._days_to_seconds(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        (0, 0),
        (60, 1),
    ),
)
def test_seconds_to_minutes(pol_info, val, expected):
    assert pol_info._seconds_to_minutes(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        (0, 0),
        (1, 60),
    ),
)
def test_minutes_to_seconds(pol_info, val, expected):
    assert pol_info._minutes_to_seconds(val) == expected


def test_strip_quotes(pol_info):
    assert pol_info._strip_quotes('"spongebob"') == "spongebob"


def test_add_quotes(pol_info):
    assert pol_info._add_quotes("squarepants") == '"squarepants"'


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        (chr(0), "Disabled"),
        (chr(1), "Enabled"),
        (chr(2), f"Invalid Value: {chr(2)!r}"),
        ("patrick", "Invalid Value"),
    ),
)
def test_binary_enable_zero_disable_one_conversion(pol_info, val, expected):
    assert pol_info._binary_enable_zero_disable_one_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, None),
        ("Disabled", chr(0)),
        ("Enabled", chr(1)),
        ("Junk", None),
    ),
)
def test_binary_enable_zero_disable_one_reverse_conversion(pol_info, val, expected):
    assert pol_info._binary_enable_zero_disable_one_reverse_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        ("0", "Administrators"),
        (0, "Administrators"),
        ("", "Administrators"),
        ("1", "Administrators and Power Users"),
        (1, "Administrators and Power Users"),
        ("2", "Administrators and Interactive Users"),
        (2, "Administrators and Interactive Users"),
        (3, "Not Defined"),
    ),
)
def test_dasd_conversion(pol_info, val, expected):
    assert pol_info._dasd_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        ("Administrators", "0"),
        ("Administrators and Power Users", "1"),
        ("Administrators and Interactive Users", "2"),
        ("Not Defined", "9999"),
        ("Plankton", "Invalid Value"),
    ),
)
def test_dasd_reverse_conversion(pol_info, val, expected):
    assert pol_info._dasd_reverse_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("Not Defined", True),
        (None, False),
        (1, True),
        (3, False),
        ("spongebob", False),
    ),
)
def test_in_range_inclusive(pol_info, val, expected):
    assert pol_info._in_range_inclusive(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        ("3,1,2", "Not Defined"),
        ("3,0", "Silently Succeed"),
        ("3,1", "Warn but allow installation"),
        ("3,2", "Do not allow installation"),
        ("3,Not Defined", "Not Defined"),
        ("3,spongebob", "Invalid Value"),
    ),
)
def test_driver_signing_reg_conversion(pol_info, val, expected):
    assert pol_info._driver_signing_reg_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Defined"),
        ("Silently Succeed", "3,0"),
        ("Warn but allow installation", f"3,{chr(1)}"),
        ("Do not allow installation", f"3,{chr(2)}"),
        ("spongebob", "Invalid Value"),
    ),
)
def test_driver_signing_reg_reverse_conversion(pol_info, val, expected):
    assert pol_info._driver_signing_reg_reverse_conversion(val) == expected


# For the next 3 tests we can't use the parametrized decorator because the
# decorator is evaluated before the imports happen, so the HAS_WIN32 is ignored
# and the decorator tries to evaluate the win32security library on systems
# without pyWin32
def test_sidConversion_no_conversion(pol_info):
    val = ws.ConvertStringSidToSid("S-1-5-0")
    expected = ["S-1-5-0"]
    assert pol_info._sidConversion([val]) == expected


def test_sidConversion_sid(pol_info):
    val = ws.ConvertStringSidToSid("S-1-1-0")
    expected = ["Everyone"]
    assert pol_info._sidConversion([val]) == expected


def test_sidConversion_name(pol_info):
    val = ws.LookupAccountName("", "DefaultAccount")[0]
    expected = [f"{socket.gethostname()}\\DefaultAccount"]
    assert pol_info._sidConversion([val]) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, None),
        ("", ""),
    ),
)
def test_usernamesToSidObjects_empty_value(pol_info, val, expected):
    assert pol_info._usernamesToSidObjects(val) == expected


def test_usernamesToSidObjects_string_list(pol_info):
    val = "DefaultAccount,Guest"
    admin_sid = ws.LookupAccountName("", "DefaultAccount")[0]
    guest_sid = ws.LookupAccountName("", "Guest")[0]
    expected = [admin_sid, guest_sid]
    assert pol_info._usernamesToSidObjects(val) == expected


def test_usernamesToSidObjects_string_list_error(pol_info):
    val = "spongebob,squarepants"
    with pytest.raises(CommandExecutionError):
        pol_info._usernamesToSidObjects(val)


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, "Not Configured"),
        ("None", "Not Configured"),
        ("true", "Run Windows PowerShell scripts first"),
        ("false", "Run Windows PowerShell scripts last"),
        ("spongebob", "Invalid Value"),
    ),
)
def test_powershell_script_order_conversion(pol_info, val, expected):
    assert pol_info._powershell_script_order_conversion(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("Not Configured", None),
        ("Run Windows PowerShell scripts first", "true"),
        ("Run Windows PowerShell scripts last", "false"),
        ("spongebob", "Invalid Value"),
    ),
)
def test_powershell_script_order_reverse_conversion(pol_info, val, expected):
    assert pol_info._powershell_script_order_reverse_conversion(val) == expected


def test_dict_lookup(pol_info):
    lookup = {
        "spongebob": "squarepants",
        "patrick": "squidward",
        "plankton": "mr.crabs",
    }
    assert pol_info._dict_lookup("spongebob", lookup=lookup) == "squarepants"
    assert (
        pol_info._dict_lookup("squarepants", lookup=lookup, value_lookup=True)
        == "spongebob"
    )
    assert pol_info._dict_lookup("homer", lookup=lookup) == "Invalid Value"
    assert (
        pol_info._dict_lookup("homer", lookup=lookup, value_lookup=True)
        == "Invalid Value"
    )
    assert pol_info._dict_lookup("homer") == "Invalid Value"


def test_dict_lookup_bitwise_add(pol_info):
    lookup = {
        0: "spongebob",
        1: "squarepants",
        2: "patrick",
    }
    assert pol_info._dict_lookup_bitwise_add("Not Defined") is None
    assert (
        pol_info._dict_lookup_bitwise_add("not a list", value_lookup=True)
        == "Invalid Value: Not a list"
    )
    assert (
        pol_info._dict_lookup_bitwise_add([], value_lookup=True)
        == "Invalid Value: No lookup passed"
    )
    assert (
        pol_info._dict_lookup_bitwise_add("not an int") == "Invalid Value: Not an int"
    )
    assert pol_info._dict_lookup_bitwise_add(0, lookup=lookup) == []
    assert (
        pol_info._dict_lookup_bitwise_add(
            ["spongebob", "squarepants"], lookup=lookup, value_lookup=True
        )
        == 1
    )
    assert pol_info._dict_lookup_bitwise_add(1, lookup=lookup) == ["squarepants"]
    assert pol_info._dict_lookup_bitwise_add(0, lookup=lookup) == []
    assert pol_info._dict_lookup_bitwise_add(0, lookup=lookup, test_zero=True) == [
        "spongebob"
    ]


@pytest.mark.parametrize(
    "val, expected",
    (
        (["list", "of", "items"], ["list", "of", "items"]),
        ("Not Defined", None),
        ("list,of,items", ["list", "of", "items"]),
        (7, "Invalid Value"),
    ),
)
def test_multi_string_put_transform(pol_info, val, expected):
    assert pol_info._multi_string_put_transform(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        (["list", "of", "items"], ["list", "of", "items"]),
        (None, "Not Defined"),
        ("list,of,items", "Invalid Value"),
        (7, "Invalid Value"),
    ),
)
def test_multi_string_get_transform(pol_info, val, expected):
    assert pol_info._multi_string_get_transform(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    (
        ("String Item", "String Item"),
        ("Not Defined", None),
        (7, None),
    ),
)
def test_string_put_transform(pol_info, val, expected):
    assert pol_info._string_put_transform(val) == expected


def test__virtual__(pol_info):
    assert win_lgpo.__virtual__() == "lgpo"
    with patch("salt.utils.platform.is_windows", return_value=False):
        assert win_lgpo.__virtual__() == (
            False,
            "win_lgpo: Not a Windows System",
        )

    with patch.object(win_lgpo, "HAS_WINDOWS_MODULES", False):
        assert win_lgpo.__virtual__() == (
            False,
            "win_lgpo: Required modules failed to load",
        )


@pytest.mark.parametrize(
    "val, expected",
    (
        (None, b"\x00\x00"),
        ("spongebob", b"s\x00p\x00o\x00n\x00g\x00e\x00b\x00o\x00b\x00\x00\x00"),
    ),
)
def test_encode_string(val, expected):
    assert win_lgpo._encode_string(val) == expected


def test_encode_string_error():
    with pytest.raises(TypeError):
        win_lgpo._encode_string(1)



"""
Unit tests for the PGJsonb returner (pgjsonb).
"""

import pytest

import salt.returners.pgjsonb as pgjsonb
from tests.support.mock import MagicMock, call, patch

if pgjsonb.HAS_PG:
    import psycopg2
    import psycopg2.extras


@pytest.fixture
def configure_loader_modules():
    return {pgjsonb: {"__opts__": {"keep_jobs_seconds": 3600, "archive_jobs": 0}}}


def test_clean_old_jobs_purge():
    """
    Tests that the function returns None when no jid_root is found.
    """
    connect_mock = MagicMock()
    with patch.object(pgjsonb, "_get_serv", connect_mock):
        with patch.dict(pgjsonb.__salt__, {"config.option": MagicMock()}):
            assert pgjsonb.clean_old_jobs() is None


def test_clean_old_jobs_archive():
    """
    Tests that the function returns None when no jid_root is found.
    """
    connect_mock = MagicMock()
    with patch.object(pgjsonb, "_get_serv", connect_mock):
        with patch.dict(pgjsonb.__salt__, {"config.option": MagicMock()}):
            with patch.dict(pgjsonb.__opts__, {"archive_jobs": 1}):
                assert pgjsonb.clean_old_jobs() is None


@pytest.mark.skipif(not pgjsonb.HAS_PG, reason="psycopg2 not installed")
def test_returner_with_bytes():
    ret = {
        "success": True,
        "return": b"bytes",
        "retcode": 0,
        "jid": "20221101172203459989",
        "fun": "file.read",
        "fun_args": ["/fake/path", {"binary": True}],
        "id": "minion-1",
    }
    decoded_ret = {
        "success": True,
        "return": "bytes",
        "retcode": 0,
        "jid": "20221101172203459989",
        "fun": "file.read",
        "fun_args": ["/fake/path", {"binary": True}],
        "id": "minion-1",
    }
    calls = [call("bytes"), call(decoded_ret)]
    with patch.object(pgjsonb, "_get_serv"):
        with patch.object(psycopg2.extras, "Json") as json_mock:
            pgjsonb.returner(ret)
            json_mock.assert_has_calls(calls)


@pytest.mark.skipif(not pgjsonb.HAS_PG, reason="psycopg2 not installed")
def test_save_load_with_bytes():
    load = {
        "return": b"bytes",
        "jid": "20221101172203459989",
    }
    decoded_load = {
        "return": "bytes",
        "jid": "20221101172203459989",
    }
    with patch.object(pgjsonb, "_get_serv"):
        with patch.object(psycopg2.extras, "Json") as json_mock:
            pgjsonb.save_load(load["jid"], load)
            json_mock.assert_called_with(decoded_load)



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

import salt.modules.mac_sysctl as mac_sysctl
from salt.exceptions import CommandExecutionError
from tests.support.mock import DEFAULT, MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {mac_sysctl: {}}


def test_get():
    """
    Tests the return of get function
    """
    mock_cmd = MagicMock(return_value="foo")
    with patch.dict(mac_sysctl.__salt__, {"cmd.run": mock_cmd}):
        assert mac_sysctl.get("kern.ostype") == "foo"


def test_assign_cmd_failed():
    """
    Tests if the assignment was successful or not
    """
    cmd = {
        "pid": 3548,
        "retcode": 1,
        "stderr": "",
        "stdout": "net.inet.icmp.icmplim: 250 -> 50",
    }
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(mac_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
        pytest.raises(
            CommandExecutionError, mac_sysctl.assign, "net.inet.icmp.icmplim", 50
        )


def test_assign():
    """
    Tests the return of successful assign function
    """
    cmd = {
        "pid": 3548,
        "retcode": 0,
        "stderr": "",
        "stdout": "net.inet.icmp.icmplim: 250 -> 50",
    }
    ret = {"net.inet.icmp.icmplim": "50"}
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(mac_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert mac_sysctl.assign("net.inet.icmp.icmplim", 50) == ret


def test_persist_no_conf_failure():
    """
    Tests adding of config file failure
    """
    read_data = IOError(13, "Permission denied", "/file")
    with patch("salt.utils.files.fopen", mock_open(read_data=read_data)), patch(
        "os.path.isfile", MagicMock(return_value=False)
    ):
        pytest.raises(
            CommandExecutionError,
            mac_sysctl.persist,
            "net.inet.icmp.icmplim",
            50,
            config=None,
        )


def test_persist_no_conf_success():
    """
    Tests successful add of config file when it did not already exist
    """
    config = "/etc/sysctl.conf"
    isfile_mock = MagicMock(side_effect=lambda x: False if x == config else DEFAULT)
    with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
        "os.path.isfile", isfile_mock
    ):
        mac_sysctl.persist("net.inet.icmp.icmplim", 50, config=config)
        num_handles = len(m_open.filehandles)
        assert num_handles == 1, num_handles
        writes = m_open.write_calls()
        num_writes = len(writes)
        assert num_writes == 1, num_writes
        assert writes[0] == "#\n# Kernel sysctl configuration\n#\n", writes[0]


def test_persist_success():
    """
    Tests successful write to existing sysctl file
    """
    config = "/etc/sysctl.conf"
    to_write = "#\n# Kernel sysctl configuration\n#\n"
    writelines_calls = [
        [
            "#\n",
            "# Kernel sysctl configuration\n",
            "#\n",
            "net.inet.icmp.icmplim=50\n",
        ]
    ]
    isfile_mock = MagicMock(side_effect=lambda x: True if x == config else DEFAULT)
    with patch(
        "salt.utils.files.fopen", mock_open(read_data=to_write)
    ) as m_open, patch("os.path.isfile", isfile_mock):
        mac_sysctl.persist("net.inet.icmp.icmplim", 50, config=config)
        num_handles = len(m_open.filehandles)
        assert num_handles == 1, num_handles
        writes = m_open.writelines_calls()
        assert writes == writelines_calls, writes



"""
The caller module is used as a front-end to manage direct calls to the salt
minion modules.
"""

import logging
import os
import sys
import traceback

import salt
import salt.channel.client
import salt.defaults.exitcodes
import salt.loader
import salt.minion
import salt.output
import salt.payload
import salt.utils.args
import salt.utils.files
import salt.utils.jid
import salt.utils.minion
import salt.utils.profile
import salt.utils.stringutils
from salt._logging import LOG_LEVELS
from salt.exceptions import (
    CommandExecutionError,
    CommandNotFoundError,
    SaltClientError,
    SaltInvocationError,
)

log = logging.getLogger(__name__)


class Caller:
    """
    Factory class to create salt-call callers for different transport
    """

    @staticmethod
    def factory(opts, **kwargs):
        return ZeroMQCaller(opts, **kwargs)


class BaseCaller:
    """
    Base class for caller transports
    """

    def __init__(self, opts):
        """
        Pass in command line opts
        """
        self.opts = opts
        self.opts["caller"] = True
        # Handle this here so other deeper code which might
        # be imported as part of the salt api doesn't do  a
        # nasty sys.exit() and tick off our developer users
        try:
            if self.opts.get("proxyid"):
                self.minion = salt.minion.SProxyMinion(opts)
            else:
                self.minion = salt.minion.SMinion(opts)
        except SaltClientError as exc:
            raise SystemExit(str(exc))

    def print_docs(self):
        """
        Pick up the documentation for all of the modules and print it out.
        """
        docs = {}
        for name, func in self.minion.functions.items():
            if name not in docs:
                if func.__doc__:
                    docs[name] = func.__doc__
        for name in sorted(docs):
            if name.startswith(self.opts.get("fun", "")):
                salt.utils.stringutils.print_cli(f"{name}:\n{docs[name]}\n")

    def print_grains(self):
        """
        Print out the grains
        """
        grains = self.minion.opts.get("grains") or salt.loader.grains(self.opts)
        salt.output.display_output({"local": grains}, "grains", self.opts)

    def run(self):
        """
        Execute the salt call logic
        """
        profiling_enabled = self.opts.get("profiling_enabled", False)
        try:
            pr = salt.utils.profile.activate_profile(profiling_enabled)
            try:
                ret = self.call()
            finally:
                salt.utils.profile.output_profile(
                    pr,
                    stats_path=self.opts.get("profiling_path", "/tmp/stats"),
                    stop=True,
                )
            out = ret.get("out", "nested")
            if self.opts["print_metadata"]:
                print_ret = ret
                out = "nested"
            else:
                print_ret = ret.get("return", {})
            salt.output.display_output(
                {"local": print_ret},
                out=out,
                opts=self.opts,
                _retcode=ret.get("retcode", 0),
            )
            # _retcode will be available in the kwargs of the outputter function
            if self.opts.get("retcode_passthrough", False):
                sys.exit(ret["retcode"])
            elif ret.get("retcode") != salt.defaults.exitcodes.EX_OK:
                sys.exit(salt.defaults.exitcodes.EX_GENERIC)
        except SaltInvocationError as err:
            raise SystemExit(err)

    def call(self):
        """
        Call the module
        """
        ret = {}
        fun = self.opts["fun"]
        ret["jid"] = salt.utils.jid.gen_jid(self.opts)
        proc_fn = os.path.join(
            salt.minion.get_proc_dir(self.opts["cachedir"]), ret["jid"]
        )
        if fun not in self.minion.functions:
            docs = self.minion.functions["sys.doc"](f"{fun}*")
            if docs:
                docs[fun] = self.minion.functions.missing_fun_string(fun)
                ret["out"] = "nested"
                ret["return"] = docs
                return ret
            sys.stderr.write(self.minion.functions.missing_fun_string(fun))
            mod_name = fun.split(".")[0]
            if mod_name in self.minion.function_errors:
                sys.stderr.write(
                    " Possible reasons: {}\n".format(
                        self.minion.function_errors[mod_name]
                    )
                )
            else:
                sys.stderr.write("\n")
            sys.exit(-1)
        metadata = self.opts.get("metadata")
        if metadata is not None:
            metadata = salt.utils.args.yamlify_arg(metadata)
        try:
            sdata = {
                "fun": fun,
                "pid": os.getpid(),
                "jid": ret["jid"],
                "tgt": "salt-call",
            }
            if metadata is not None:
                sdata["metadata"] = metadata
            args, kwargs = salt.minion.load_args_and_kwargs(
                self.minion.functions[fun],
                salt.utils.args.parse_input(
                    self.opts["arg"], no_parse=self.opts.get("no_parse", [])
                ),
                data=sdata,
            )
            try:
                with salt.utils.files.fopen(proc_fn, "w+b") as fp_:
                    fp_.write(salt.payload.dumps(sdata))
            except NameError:
                # Don't require msgpack with local
                pass
            except OSError:
                sys.stderr.write(
                    "Cannot write to process directory. "
                    "Do you have permissions to "
                    "write to {} ?\n".format(proc_fn)
                )
            func = self.minion.functions[fun]
            data = {"arg": args, "fun": fun}
            data.update(kwargs)
            executors = getattr(
                self.minion, "module_executors", []
            ) or salt.utils.args.yamlify_arg(
                self.opts.get("module_executors", "[direct_call]")
            )
            if self.opts.get("executor_opts", None):
                data["executor_opts"] = salt.utils.args.yamlify_arg(
                    self.opts["executor_opts"]
                )
            if isinstance(executors, str):
                executors = [executors]
            try:
                for name in executors:
                    fname = f"{name}.execute"
                    if fname not in self.minion.executors:
                        raise SaltInvocationError(f"Executor '{name}' is not available")
                    ret["return"] = self.minion.executors[fname](
                        self.opts, data, func, args, kwargs
                    )
                    if ret["return"] is not None:
                        break
            except TypeError as exc:
                sys.stderr.write(f"\nPassed invalid arguments: {exc}.\n\nUsage:\n")
                salt.utils.stringutils.print_cli(func.__doc__)
                active_level = LOG_LEVELS.get(
                    self.opts["log_level"].lower(), logging.ERROR
                )
                if active_level <= logging.DEBUG:
                    trace = traceback.format_exc()
                    sys.stderr.write(trace)
                sys.exit(salt.defaults.exitcodes.EX_GENERIC)
            try:
                retcode = self.minion.executors.pack["__context__"].get("retcode", 0)
            except AttributeError:
                retcode = salt.defaults.exitcodes.EX_GENERIC

            if retcode == 0:
                # No nonzero retcode in __context__ dunder. Check if return
                # is a dictionary with a "result" or "success" key.
                try:
                    func_result = all(
                        ret["return"].get(x, True) for x in ("result", "success")
                    )
                except Exception:  # pylint: disable=broad-except
                    # return data is not a dict
                    func_result = True
                if not func_result:
                    retcode = salt.defaults.exitcodes.EX_GENERIC

            ret["retcode"] = retcode
        except CommandExecutionError as exc:
            msg = "Error running '{0}': {1}\n"
            active_level = LOG_LEVELS.get(self.opts["log_level"].lower(), logging.ERROR)
            if active_level <= logging.DEBUG:
                sys.stderr.write(traceback.format_exc())
            sys.stderr.write(msg.format(fun, exc))
            sys.exit(salt.defaults.exitcodes.EX_GENERIC)
        except CommandNotFoundError as exc:
            msg = "Command required for '{0}' not found: {1}\n"
            sys.stderr.write(msg.format(fun, exc))
            sys.exit(salt.defaults.exitcodes.EX_GENERIC)
        try:
            os.remove(proc_fn)
        except OSError:
            pass
        if hasattr(self.minion.functions[fun], "__outputter__"):
            oput = self.minion.functions[fun].__outputter__
            if isinstance(oput, str):
                ret["out"] = oput
        is_local = (
            self.opts["local"]
            or self.opts.get("file_client", False) == "local"
            or self.opts.get("master_type") == "disable"
        )
        returners = self.opts.get("return", "").split(",")
        if (not is_local) or returners:
            ret["id"] = self.opts["id"]
            ret["fun"] = fun
            ret["fun_args"] = self.opts["arg"]
            if metadata is not None:
                ret["metadata"] = metadata

        for returner in returners:
            if not returner:  # if we got an empty returner somehow, skip
                continue
            try:
                ret["success"] = True
                self.minion.returners[f"{returner}.returner"](ret)
            except Exception:  # pylint: disable=broad-except
                pass

        # return the job infos back up to the respective minion's master
        if not is_local and not self.opts.get("no_return_event", False):
            try:
                mret = ret.copy()
                mret["jid"] = "req"
                self.return_pub(mret)
            except Exception:  # pylint: disable=broad-except
                pass
        elif self.opts["cache_jobs"]:
            # Local job cache has been enabled
            salt.utils.minion.cache_jobs(self.opts, ret["jid"], ret)

        return ret


class ZeroMQCaller(BaseCaller):
    """
    Object to wrap the calling of local salt modules for the salt-call command
    """

    def __init__(self, opts):  # pylint: disable=useless-super-delegation
        """
        Pass in the command line options
        """
        super().__init__(opts)

    def return_pub(self, ret):
        """
        Return the data up to the master
        """
        with salt.channel.client.ReqChannel.factory(
            self.opts, usage="salt_call"
        ) as channel:
            load = {"cmd": "_return", "id": self.opts["id"]}
            for key, value in ret.items():
                load[key] = value
            channel.send(load)
